# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "ABS_TOL",
    "REL_TOL",
    "MIN_LENGTH",
    "MAX_LENGTH",
    "ABS_TOL_ANGLE",
    "REL_TOL_ANGLE",
    "MIN_ANGLE",
    "ATOM_TYPES",
    "PEARSON_SYMBOLS",
    "BRAVAIS_LATTICE_NAMES",
    "BRAVAIS_LATTICE_VARIATIONS",
    "TRANSFORM_TO_CONVENTIONAL",
    "DEFAULT_K_PATHS",
    "HS_PLOT_NAMES",
]

# Length variables
ABS_TOL = 1e-8  # Meant for the linear spatial variables
REL_TOL = 1e-4  # Meant for the linear spatial variables
# MIN_LENGTH is a direct consequence of the REL_TOL and ABS_TOL:
# for l = MIN_LENGTH => ABS_TOL = l * REL_TOL
MIN_LENGTH = ABS_TOL / REL_TOL
# MAX_LENGTH is a direct consequence of the ABS_TOL:
# Inverse of the MAX_LENGTH in the real space has to be meaningful
# in the reciprocal space (< ABS_TOL).
MAX_LENGTH = 1 / ABS_TOL

# TODO Think how to connect angle tolerance with spatial tolerance.

ABS_TOL_ANGLE = 1e-4  # Meant for the angular variables, in degrees.
REL_TOL_ANGLE = 1e-2  # Meant for the angular variables.
# MIN_ANGLE is a direct consequence of the REL_TOL_ANGLE and ABS_TOL_ANGLE:
# for a = MIN_ANGLE => ABS_TOL_ANGLE = a * REL_TOL_ANGLE
MIN_ANGLE = ABS_TOL_ANGLE / REL_TOL_ANGLE  # In degrees

ATOM_TYPES = (
    "H",
    "He",
    "Li",
    "Be",
    "B",
    "C",
    "N",
    "O",
    "F",
    "Ne",
    "Na",
    "Mg",
    "Al",
    "Si",
    "P",
    "S",
    "Cl",
    "Ar",
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Ge",
    "As",
    "Se",
    "Br",
    "Kr",
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Sb",
    "Te",
    "I",
    "Xe",
    "Cs",
    "Ba",
    "La",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
    "Po",
    "At",
    "Rn",
    "Fr",
    "Ra",
    "Ac",
    "Th",
    "Pa",
    "U",
    "Np",
    "Pu",
    "Am",
    "Cm",
    "Bk",
    "Cf",
    "Es",
    "Fm",
    "Md",
    "No",
    "Lr",
    "Rf",
    "Db",
    "Sg",
    "Bh",
    "Hs",
    "Mt",
    "Ds",
    "Rg",
    "Cn",
    "Nh",
    "Fl",
    "Mc",
    "Lv",
    "Ts",
    "Og",
)


PEARSON_SYMBOLS = {
    "CUB": "cP",
    "FCC": "cF",
    "BCC": "cI",
    "TET": "tP",
    "BCT": "tI",
    "ORC": "oP",
    "ORCF": "oF",
    "ORCI": "oI",
    "ORCC": "oS",
    "HEX": "hP",
    "RHL": "hR",
    "MCL": "mP",
    "MCLC": "mS",
    "TRI": "aP",
}

BRAVAIS_LATTICE_NAMES = {
    "CUB": "Cubic",
    "FCC": "Face-centered cubic",
    "BCC": "Body-centered cubic",
    "TET": "Tetragonal",
    "BCT": "Body-centered tetragonal",
    "ORC": "Orthorhombic",
    "ORCF": "Face-centered orthorhombic",
    "ORCI": "Body-centered orthorhombic",
    "ORCC": "C-centered orthorhombic",
    "HEX": "Hexagonal",
    "RHL": "Rhombohedral",
    "MCL": "Monoclinic",
    "MCLC": "C-centered monoclinic",
    "TRI": "Triclinic",
}

BRAVAIS_LATTICE_VARIATIONS = [
    "CUB",
    "FCC",
    "BCC",
    "TET",
    "BCT1",
    "BCT2",
    "ORC",
    "ORCF1",
    "ORCF2",
    "ORCF3",
    "ORCI",
    "ORCC",
    "HEX",
    "RHL1",
    "RHL2",
    "MCL",
    "MCLC1",
    "MCLC2",
    "MCLC3",
    "MCLC4",
    "MCLC5",
    "TRI1a",
    "TRI2a",
    "TRI1b",
    "TRI2b",
]

TRANSFORM_TO_CONVENTIONAL = {
    "CUB": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    "FCC": [[-1.0, 1.0, 1.0], [1.0, -1.0, 1.0], [1.0, 1.0, -1.0]],
    "BCC": [[0, 1.0, 1.0], [1.0, 0, 1.0], [1.0, 1.0, 0]],
    "TET": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    "BCT": [[0, 1.0, 1.0], [1.0, 0, 1.0], [1.0, 1.0, 0]],
    "ORC": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    "ORCF": [[-1.0, 1.0, 1.0], [1.0, -1.0, 1.0], [1.0, 1.0, -1.0]],
    "ORCI": [[0, 1.0, 1.0], [1.0, 0, 1.0], [1.0, 1.0, 0]],
    "ORCC": [[1.0, 1.0, 0], [-1.0, 1.0, 0], [0, 0, 1.0]],
    "HEX": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    "RHL": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    "MCL": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    "MCLC": [[1.0, -1.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    "TRI": [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
}

DEFAULT_K_PATHS = {
    "CUB": "G-X-M-G-R-X|M-R",
    "FCC": "G-X-W-K-G-L-U-W-L-K|U-X",
    "BCC": "G-H-N-G-P-H|P-N",
    "TET": "G-X-M-G-Z-R-A-Z|X-R|M-A",
    "BCT1": "G-X-M-G-Z-P-N-Z1-M|X-P",
    "BCT2": "G-X-Y-S-G-Z-S1-N-P-Y1-Z|X-P",
    "ORC": "G-X-S-Y-G-Z-U-R-T-Z|Y-T|U-X|S-R",
    "ORCF1": "G-Y-T-Z-G-X-A1-Y|T-X1|X-A-Z|L-G",
    "ORCF2": "G-Y-C-D-X-G-Z-D1-H-C|C1-Z|X-H1|H-Y|L-G",
    "ORCF3": "G-Y-T-Z-G-X-A1-Y|X-A-Z|L-G",
    "ORCI": "G-X-L-T-W-R-X1-Z-G-Y-S-W|L1-Y|Y1-Z",
    "ORCC": "G-X-S-R-A-Z-G-Y-X1-A1-T-Y|Z-T",
    "HEX": "G-M-K-G-A-L-H-A|L-M|K-H",
    "RHL1": "G-L-B1|B-Z-G-X|Q-F-P1-Z|L-P",
    "RHL2": "G-P-Z-Q-G-F-P1-Q1-L-Z",
    "MCL": "G-Y-H-C-E-M1-A-X-H1|M-D-Z|Y-D",
    "MCLC1": "G-Y-F-L-I|I1-Z-F1|Y-X1|X-G-N|M-G",
    "MCLC2": "G-Y-F-L-I|I1-Z-F1|N-G-M",
    "MCLC3": "G-Y-F-H-Z-I-F1|H1-Y1-X-G-N|M-G",
    "MCLC4": "G-Y-F-H-Z-I|H1-Y1-X-G-N|M-G",
    "MCLC5": "G-Y-F-L-I|I1-Z-H-F1|H1-Y1-X-G-N|M-G",
    "TRI1a": "X-G-Y|L-G-Z|N-G-M|R-G",
    "TRI1b": "X-G-Y|L-G-Z|N-G-M|R-G",
    "TRI2a": "X-G-Y|L-G-Z|N-G-M|R-G",
    "TRI2b": "X-G-Y|L-G-Z|N-G-M|R-G",
}

HS_PLOT_NAMES = {
    "G": "$\\Gamma$",
    "M": "M",
    "R": "R",
    "X": "X",
    "K": "K",
    "L": "L",
    "U": "U",
    "W": "W",
    "H": "H",
    "P": "P",
    "N": "N",
    "A": "A",
    "Z": "Z",
    "Z1": "Z$_1$",
    "Y": "Y",
    "Y1": "Y$_1$",
    "S": "S",  # it is overwritten to sigma if needed.
    "S1": "S$_1$",  # it is overwritten to sigma if needed.
    "T": "T",
    "A1": "A$_1$",
    "X1": "X$_1$",
    "C": "C",
    "C1": "C$_1$",
    "D": "D",
    "D1": "D$_1$",
    "H1": "H$_1$",
    "L1": "L$_1$",
    "L2": "L$_2$",
    "B": "B",
    "B1": "B$_1$",
    "F": "F",
    "P1": "P$_1$",
    "P2": "P$_2$",
    "Q": "Q",
    "Q1": "Q$_1$",
    "E": "E",
    "H2": "H$_2$",
    "M1": "M$_1$",
    "M2": "M$_2$",
    "N1": "N$_1$",
    "F1": "F$_1$",
    "F2": "F$_2$",
    "F3": "F$_3$",
    "I": "I",
    "I1": "I$_1$",
    "X2": "X$_2$",
    "Y2": "Y$_2$",
    "Y3": "Y$_3$",
}



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Exchange template.

Write a tutorial with docstring here.
"""
__all__ = ["ExchangeTemplate"]


class ExchangeTemplate:
    r"""
    Store a template for the :py:class:`.SpinHamiltonian`.

    In addition stores the technical details for plotting,
    orbital decomposition, etc.

    Attributes
    ----------
    names : dict
        Dictionary of neighbors from the template file.

        .. code-block:: python

            {name : [(atom1, atom2, R), ...], ...}

    latex_names : dict
        The dictionary of Latex version of names from `names`.

        .. code-block:: python

            {name : latex_name, ...}
    """

    def __init__(self) -> None:
        self.names = {}
        self.latex_names = {}

    def get_list(self):
        r"""
        Translate bonds present in the template into the list.

        Returns
        -------
        template_list : list
            List with the bond specifications:

            .. code-block:: python

                [(atom_1, atom_2, R), ...]
        """
        template_list = []
        for name in self.names:
            for atom1, atom2, R in self.names[name]:
                template_list.append((atom1, atom2, R))
        return template_list



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Module for density of states post-processing.
"""

from .dos import DOSQE
from .fatbands_plotting import *
from .pdos import PDOS, PDOSQE
from .pdos_plotting import *
from .plotting import *

__all__ = ["DOSQE", "PDOSQE", "PDOS"]
__all__.extend(plotting.__all__)
__all__.extend(fatbands_plotting.__all__)
__all__.extend(pdos_plotting.__all__)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from math import floor, log10
from typing import Union

import numpy as np

import radtools.crystal.cell as Cell
from radtools.crystal.atom import Atom
from radtools.crystal.lattice import Lattice
from radtools.crystal.properties import dipole_dipole_energy, dipole_dipole_interaction
from radtools.geometry import absolute_to_relative


class Crystal(Lattice):
    r"""
    Crystal class.

    It is a child of Lattice class.

    Iterable over atoms. All attributes of the :py:class:`.Lattice`
    are accessible directly from the crystal:

    .. doctest::

        >>> import radtools as rad
        >>> cub = rad.lattice_example("CUB")
        >>> crystal = rad.Crystal(cub)
        >>> crystal.pearson_symbol
        'cP'

    For the full description of the lattice attributes and methods
    see :ref:`guide_crystal_lattice`.

    Parameters
    ----------
    lattice : :py:class:`.Lattice`, optional
        Lattice of the crystal. If not provided,
        then orthonormal lattice is used ("CUB with :math:`a = 1`).
    atoms : list, optional
        List of :py:class:`Atom` objects.
    relative : bool, default True
        Whether ``atoms`` positions are in relative coordinates.
    standardize : bool, default True
        Whether to standardize the lattice.
    **kwargs
        Keyword arguments for :py:class:`.Lattice` initialization.

    Attributes
    ----------
    atoms : list
        List of atoms of the crystal.
    """

    def __init__(
        self,
        lattice: Lattice = None,
        atoms=None,
        relative=True,
        standardize=True,
        **kwargs,
    ) -> None:
        self.atoms = []
        if lattice is None:
            if len(kwargs) == 0:
                kwargs["cell"] = [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
        else:
            kwargs = {}
            kwargs["cell"] = lattice.cell

        super().__init__(standardize=standardize, **kwargs)

        if atoms is not None:
            for a in atoms:
                self.add_atom(a, relative=relative)

    def __iter__(self):
        return CrystalIterator(self)

    def __contains__(self, atom: Union[Atom, str]):
        if isinstance(atom, str):
            try:
                atom = self.get_atom(atom, return_all=True)
                return True
            except ValueError:
                return False
        return atom in self.atoms

    def __len__(self):
        return self.atoms.__len__()

    def __getitem__(self, name) -> Atom:
        return self.get_atom(name, return_all=True)

    def __getattr__(self, name):
        # Fix copy/deepcopy RecursionError
        if name in ["__setstate__"]:
            raise AttributeError(name)
        try:
            atom = self.get_atom(name=name)
            return atom
        except ValueError:
            raise AttributeError(f"'Crystal' object has no attribute '{name}'")

    @property
    def lattice(self):
        r"""
        Lattice of the crystal.

        It returns an independent instance of the :py:class:`.Lattice` class.
        You can use it to play with the crystal`s lattice independently,
        but it will not affect the crystal itself.

        See :ref:`guide_crystal_lattice` for details.

        Returns
        -------
        lattice : :py:class:`.Lattice`
            Lattice of the crystal.

        Notes
        -----
        New :py:class:`.Lattice` object is created each time you call this property.
        It is created with ``standardize=False`` parameter.
        """
        return Lattice(self.cell, standardize=False)

    def add_atom(self, new_atom: Union[Atom, str] = None, relative=True, **kwargs):
        r"""
        Add atom to the crystal.

        If name and index of the ``new_atom`` are the same as of some atom of the crystal,
        then ``new_atom`` is not added.

        If index of ``new_atom`` is not defined, it is set.

        Parameters
        ----------
        new_atoms : :py:class:`.Atom` or str, optional
            New atom. All kwargs are ignored if ``new_atom`` is not ``None`` and of type :py:class:`.Atom`.
            If ``str``, then pair ``name : new_atom`` is added to ``kwargs``.
        relative : bool, default True
            Whether ``new_atom`` position is in relative coordinates.
        **kwargs
            Keyword arguments for :py:class:`.Atom` initialization.

        Raises
        ------
        TypeError
            If ``new_atom`` is not an :py:class:`.Atom`.
        ValueError
            If the atom is already present in the crystal.
        """

        if isinstance(new_atom, str):
            kwargs["name"] = new_atom
            new_atom = None

        if new_atom is None:
            new_atom = Atom(**kwargs)

        if not isinstance(new_atom, Atom):
            raise TypeError("New atom is not an Atom. " + f"Received {type(new_atom)}.")
        try:
            i = new_atom.index
        except ValueError:
            new_atom.index = len(self.atoms) + 1

        if not relative:
            new_atom.position = absolute_to_relative(self.cell, new_atom.position)

        if new_atom not in self.atoms:
            self.atoms.append(new_atom)
        else:
            raise ValueError("Atom is already in the crystal.")

    def remove_atom(self, atom: Union[Atom, str], index=None):
        r"""
        Remove atom from the crystal.

        If type(``atom``) == ``str``, then all atoms with the name ``atom`` are removed
        if ``index`` == ``None`` or only atom with the name ``atom`` and index ``index`` is removed.

        It type(``atom``) == :py:class:`.Atom`, ``index`` is ignored.

        Parameters
        ----------
        atom : :py:class:`.Atom` or str
            :py:class`.Atom` object or atom`s name.
            If name, then it has to be unique among atoms of the crystal.
        index : optional
            Index of the atom.

        Raises
        ------
        ValueError
            If no match is found.
        """

        if isinstance(atom, str):
            atoms = self.get_atom(atom, index=index, return_all=True)
        else:
            atoms = [atom]
        for atom in atoms:
            self.atoms.remove(atom)

    # Modification of position has to be avoided here.
    def get_atom(self, name, index=None, return_all=False):
        r"""
        Return atom object of the crystal.

        Notes
        -----
        :py:attr:`.index` in combination with :py:attr:`.name` is supposed to be a unique value,
        however it uniqueness is not strictly checked,
        pay attention in custom cases.

        Parameters
        ----------
        name : str
            Name of the atom. In general not unique. If ``name`` contains "__", then it is split
            into ``name`` and ``index``.
        index : int, optional
            Index of the atom.
        return_all : bool, default False
            Whether to return the list of non-unique matches or raise an ``ValueError``.

        Returns
        -------
        atom : :py:class:`.Atom` or list
            If only one atom is found, then :py:class:`.Atom` object is returned.
            If several atoms are found and ``return_all`` is ``True``,
            then list of :py:class:`.Atom` objects is returned.

        Raises
        ------
        ValueError
            If no match is found or the match is not unique and ``return_all`` is ``False``.
        """

        atoms = []
        if "__" in name and index is None:
            name, index = name.split("__")
            index = int(index)

        for atom in self.atoms:
            if atom.name == name:
                if index is None or atom.index == index:
                    atoms.append(atom)

        if len(atoms) == 0:
            raise ValueError(f"No match found for name = {name}, index = {index}")
        elif len(atoms) == 1:
            if return_all:
                return atoms
            return atoms[0]
        elif not return_all:
            raise ValueError(
                f"Multiple matches found for name = {name}, index = {index}"
            )
        return atoms

    def get_atom_coordinates(
        self, atom: Union[Atom, str], R=(0, 0, 0), index=None, relative=True
    ):
        r"""
        Getter for the atom coordinates.

        Parameters
        ----------
        atom : :py:class:`.Atom` or str
            :py:class`.Atom` object or atom`s name.
            If name, then it has to be unique among atoms of the crystal.
        index : int, optional
            Index of the atom.
        R : (3,) |array_like|_, default (0, 0, 0)
            Radius vector of the unit cell for atom2 (i,j,k).
        relative : bool, default True
            Whether to return relative coordinates.

        Returns
        -------
        coordinates : 1 x 3 array
            Coordinates of atom in the cell R in absolute coordinates.
        """

        if isinstance(atom, str):
            atom = self.get_atom(atom, index=index)
        elif atom not in self.atoms:
            raise ValueError(f"There is no {atom} in the crystal.")

        rel_coordinates = np.array(R + atom.position)

        if relative:
            return rel_coordinates

        return rel_coordinates @ self.cell

    def get_vector(
        self,
        atom1: Union[Atom, str],
        atom2: Union[Atom, str],
        R=(0, 0, 0),
        index1=None,
        index2=None,
        relative=False,
    ):
        r"""
        Getter for vector from atom1 to atom2.

        Parameters
        ----------
        atom1 : :py:class:`.Atom` or str
            :py:class`.Atom` object or atom`s name in (0, 0, 0) unit cell.
            If name, then it has to be unique among atoms of the crystal.
        atom2 : :py:class:`.Atom` or str
            :py:class`.Atom` object or atom`s name in ``R`` unit cell.
            If name, then it has to be unique among atoms of the crystal.
        R : (3,) |array_like|_, default (0, 0, 0)
            Radius vector of the unit cell for atom2 (i,j,k).
        relative : bool, default False
            Whether to return the vector relative coordinates.
        index1 : int, optional
            Index of the atom1.
        index2 : int, optional
            Index of the atom2.

        Returns
        -------
        v : (3,) :numpy:`ndarray`
            Vector from atom1 in (0,0,0) cell to atom2 in R cell.
        """

        coord1 = self.get_atom_coordinates(atom1, index=index1, relative=relative)
        coord2 = self.get_atom_coordinates(atom2, R, index=index2, relative=relative)

        return coord2 - coord1

    def get_distance(
        self,
        atom1: Union[Atom, str],
        atom2: Union[Atom, str],
        R=(0, 0, 0),
        index1=None,
        index2=None,
        relative=False,
    ):
        r"""
        Getter for distance between the atom1 and atom2.

        Parameters
        ----------
        atom1 : :py:class:`.Atom` or str
            :py:class`.Atom` object or atom`s name in (0, 0, 0) unit cell.
            If name, then it has to be unique among atoms of the crystal.
        atom2 : :py:class:`.Atom` or str
            :py:class`.Atom` object or atom`s name in ``R`` unit cell.
            If name, then it has to be unique among atoms of the crystal.
        R : (3,) |array_like|_, default (0, 0, 0)
            Radius vector of the unit cell for atom2 (i,j,k).
        relative : bool, default False
            Whether to use relative coordinates. (Strange, but if you wish)
        index1 : int, optional
            Index of the atom1.
        index2 : int, optional
            Index of the atom2.

        Returns
        -------
        distance : floats
            Distance between atom1 in (0,0,0) cell and atom2 in R cell.
        """

        return np.linalg.norm(
            self.get_vector(
                atom1, atom2, R, index1=index1, index2=index2, relative=relative
            )
        )

    def find_primitive_cell(self):
        r"""
        Detect primitive cell.

        Before the detection of the primitive cell the corresponding bravais lattice type may not
        be correct, since it is determined with the current cell, which is not necessary primitive one.
        """
        self.cell, self.atoms = Cell.primitive(self.cell, self.atoms)

    def mag_dipdip_energy(self, na, nb, nc, progress_bar=True):
        r"""
        Computes magnetic dipole-dipole energy.

        .. math::

            E = -\frac{\mu_0}{4\pi}\sum_{i > j}\left(3(\vec{m_i} \cdot \vec{r_{ij}})(\vec{m_j} \cdot \vec{r_{ij}}) - (\vec{m_i}\cdot\vec{m_j})\right)

        Parameters
        ----------
        na : int
            Translations along :py:attr:`.Crystal.a1`.
        nb : int
            Translations along :py:attr:`.Crystal.a2`.
        nc : int
            Translations along :py:attr:`.Crystal.a3`.
        progress_bar : bool, default True
            Whether to show progressbar.

        Returns
        -------
        energy : float
            Dipole-dipole energy in meV.

        See Also
        --------
        dipole_dipole_energy
        """

        translations = (
            np.transpose(np.indices((na, nb, nc)), (1, 2, 3, 0)).reshape(
                (na * nb * nc, 3)
            )
            @ self.cell
        )
        n_t = len(translations)

        magnetic_atoms = []
        for atom in self.atoms:
            try:
                tmp = atom.magmom
                magnetic_atoms.append(atom)
            except ValueError:
                pass

        if len(magnetic_atoms) == 0:
            raise ValueError("There are no magnetic atoms in the crystal.")

        magnetic_centres = np.zeros((n_t * len(magnetic_atoms), 2, 3), dtype=float)
        for a_i, atom in enumerate(magnetic_atoms):
            magnetic_centres[a_i * n_t : (a_i + 1) * n_t, 0] = np.tile(
                atom.magmom, (n_t, 1)
            )
            magnetic_centres[a_i * n_t : (a_i + 1) * n_t, 1] = (
                translations + atom.position @ self.cell
            )

        return dipole_dipole_energy(magnetic_centres, progress_bar=progress_bar)

    def converge_mag_dipdip_energy(
        self,
        start=(10, 10, 10),
        step=(10, 10, 10),
        eps=10e-3,
        progress_bar=True,
        verbose=False,
    ):
        r"""
        Converge magnetic dipole-dipole energy.

        Parameters
        ----------
        start : tuple of 3 int, default (10, 10, 10)
            (na, nb, nc) starting values.
        step : tuple of 3 int, default (10, 10, 10)
            (na, nb, nc) step values. If 0, then no step is applied.
        eps : float, default 1e-3
            Convergence parameter.
        progress_bar : bool, default False
            Whether to show progressbar.
        verbose : bool, default False
            Whether to print information about each step.

        Returns
        -------
        energies : list
            List of (na,nb,nc) and energies for each step.

            .. code-block:: python

                energies = [((na, nb, nc), energy), ...]

        See Also
        --------
        dipole_dipole_energy
        dipole_dipole_interaction
        """

        na, nb, nc = start
        da, db, dc = step

        translations = (
            np.transpose(np.indices((na, nb, nc)), (1, 2, 3, 0)).reshape(
                (na * nb * nc, 3)
            )
            @ self.cell
        )

        magnetic_atoms = []
        for atom in self.atoms:
            try:
                tmp = atom.magmom
                magnetic_atoms.append(atom)
            except ValueError:
                pass
        n_atoms = len(magnetic_atoms)

        n_t = len(translations)
        mc1 = np.zeros((n_t * n_atoms, 2, 3), dtype=float)
        for a_i, atom in enumerate(magnetic_atoms):
            mc1[a_i * n_t : (a_i + 1) * n_t, 0] = np.tile(atom.magmom, (n_t, 1))
            mc1[a_i * n_t : (a_i + 1) * n_t, 1] = (
                translations + atom.position @ self.cell
            )

        energy1 = dipole_dipole_energy(mc1, progress_bar=progress_bar, normalize=False)
        energies = [(start, energy1 / len(mc1))]
        if verbose:
            n_digits = abs(floor(log10(abs(eps)))) + 2
            print(
                f"{'Size':^{6+len(str(na))+len(str(nb))+len(str(nc))}} {'Energy':^{n_digits+4}} {'Difference':^{n_digits+4}}"
            )
            print(
                f"({na}, {nb}, {nc}) "
                + f"{energy1 /len(mc1):<{n_digits+4}.{n_digits}f}"
            )

        if da == 0 and db == 0 and dc == 0:
            return energies

        difference = 10 * eps
        while difference > eps:
            translations = []
            if dc != 0:
                for k in range(nc, nc + dc):
                    for j in range(0, nb + db):
                        for i in range(0, na + da):
                            translations.append((i, j, k))
            for k in range(0, nc):
                if db != 0:
                    for j in range(nb, nb + db):
                        for i in range(0, na + da):
                            translations.append((i, j, k))
                for j in range(0, nb):
                    for i in range(na, na + da):
                        translations.append((i, j, k))
            translations = np.array(translations) @ self.cell

            n_t = len(translations)
            mc2 = np.zeros((n_t * n_atoms, 2, 3), dtype=float)
            for a_i, atom in enumerate(magnetic_atoms):
                mc2[a_i * n_t : (a_i + 1) * n_t, 0] = np.tile(atom.magmom, (n_t, 1))
                mc2[a_i * n_t : (a_i + 1) * n_t, 1] = (
                    translations + atom.position @ self.cell
                )
            energy2 = dipole_dipole_energy(
                mc2, progress_bar=progress_bar, normalize=False
            )
            energy_inter = dipole_dipole_interaction(
                mc1, mc2, progress_bar=progress_bar, normalize=False
            )
            new_energy = energy1 + energy_inter + energy2

            energy1 = new_energy
            nc += dc
            nb += db
            na += da
            energies.append(((na, nb, nc), energy1 / (len(mc2) + len(mc1))))
            difference = abs(energies[-2][1] - energies[-1][1])
            if verbose:
                print(
                    f"({na}, {nb}, {nc}) "
                    + f"{energy1 / (len(mc2) +len(mc1)):<{n_digits+4}.{n_digits}f} "
                    + f"{difference:<{n_digits+4}.{n_digits}f}"
                )

            mc1 = np.concatenate((mc1, mc2))

        return energies


class CrystalIterator:
    def __init__(self, crystal: Crystal) -> None:
        self._list = crystal.atoms
        self._index = 0

    def __next__(self) -> Atom:
        if self._index < len(self._list):
            result = self._list[self._index]
            self._index += 1
            return result
        raise StopIteration

    def __iter__(self):
        return self



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .constructor import *
from .examples import *
from .hs_points import *
from .standardize import *
from .variations import *

__all__ = []
__all__.extend(examples.__all__)
__all__.extend(standardize.__all__)
__all__.extend(constructor.__all__)
__all__.extend(hs_points.__all__)
__all__.extend(variations.__all__)


r"""
Bravais lattices.
"""

# High symmetry k points getters


# # 5
# class BCT(Lattice):
#     def __init__(self) -> None:
#         self._PLOT_NAMES["S"] = "$\\Sigma$"
#         self._PLOT_NAMES["S1"] = "$\\Sigma_1$"


# # 14
# class TRI(Lattice):
#     def __init__(
#         self,
#         a: float,
#         b: float,
#         c: float,
#         alpha: float,
#         beta: float,
#         gamma: float,
#         reciprocal=False,
#     ) -> None:
#         if not reciprocal:
#             tmp = sorted([(a, alpha), (b, beta), (c, gamma)], key=lambda x: x[0])
#             a = tmp[0][0]
#             alpha = tmp[0][1]
#             b = tmp[1][0]
#             beta = tmp[1][1]
#             c = tmp[2][0]
#             gamma = tmp[2][1]
#         super().__init__(a, b, c, alpha, beta, gamma)
#         if reciprocal:
#             self.cell = self.reciprocal_cell
#         self.conv_a = a
#         self.conv_b = b
#         self.conv_c = c
#         self.conv_alpha = alpha
#         self.conv_beta = beta
#         self.conv_gamma = gamma
#         if (
#             self.k_alpha < self.k_gamma < self.k_beta
#             or self.k_beta < self.k_gamma < self.k_alpha
#         ):
#             raise RuntimeError("k_gamma is not minimal nor maximal.")
#         self.conv_cell = self.cell



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from argparse import ArgumentParser

from termcolor import cprint

from radtools.io.internal import dump_spinham_txt, load_template
from radtools.io.tb2j import load_tb2j_model


def manager(
    input_filename,
    template_file=None,
    output_name=None,
    decimals=4,
    form_model=False,
    no_anisotropic=False,
    no_matrix=False,
    nodmi=False,
    verbose=False,
    max_distance=None,
    min_distance=None,
):
    r"""
    :ref:`rad-extract-tb2j` script.

    Full documentation on the behaviour is available in the
    :ref:`User Guide <rad-extract-tb2j>`.
    Parameters of the function directly
    correspond to the arguments of the script.

    Parameters
    ----------
    input_filename : str
        Relative or absolute path to the "exchange.out" file,
        including the name and extension of the file itself.

        Console argument: ``-if`` / ``--input-filename``
    template_file : str, optional
        Relative or absolute path to the template file,
        including the name and extension of the file.

        Console argument: ``-tf`` / ``--template-file``
    output_name : str, optional
        Name of the output files.

        If this parameter is not specified, the result is printed in the
        standard output stream.

        Console argument: ``-on`` / ``--output-name``
    decimals : int, default 4
        Decimals after the comma for the exchange values.

        Console argument: ``-d`` / ``--decimals``

        .. versionchanged:: 0.5.17 Renamed from ``-acc``/``--accuracy``.
    form_model : bool, default False
        Whether to form the model from the template.

        Console argument: ``-fm`` / ``--form-model``

        .. versionchanged:: 0.8.0 Renamed from ``-fs``/``--force-symmetry``.
    no_anisotropic : bool, default False
        Whether to output anisotropic exchange.

        Console argument: ``-noa`` / ``--no-anisotropic``

        .. versionchanged:: 0.8.0 Renamed from ``-a``/``--anisotropic``.
    no_matrix : bool, default False
        Whether to output whole matrix exchange.

        Console argument: ``-nom`` / ``--no-matrix``

        .. versionchanged:: 0.8.0 Renamed from ``-m``/``--matrix``.
    nodmi : bool, default False
        Whether to output DMI exchange.

        Console argument: ``-nodmi``

        .. versionchanged:: 0.8.0 Renamed from ``-dmi``.
    verbose : bool, default False
        Verbose output, propagates to the called methods.

        Console argument: ``-v`` / ``--verbose``
    max_distance : float, optional
        (<=) Maximum distance.

        All the bonds with the distance between atom 1 and atom 2
        greater than maximum distance are excluded from the model.

        Console argument: ``-maxd`` / ``--max-distance``

        .. versionadded:: 0.8.0
    min_distance : float, optional
        (>=) Minimum distance.

        All the bonds with the distance between atom 1 and atom 2
        lower than minimum distance are excluded from the Hamiltonian.

        Console argument: ``-mind`` / ``--min-distance``

        .. versionadded:: 0.8.0

    """

    if form_model and template_file is None:
        raise ValueError(
            "Template file is required for forming the model from the template."
            + "--form-model implies --template-file"
        )

    head, _ = os.path.split(input_filename)
    if output_name is not None:
        out_head, out_tail = os.path.split(output_name)
        if len(out_head) == 0:
            out_head = head
        if len(out_tail) == 0:
            out_tail = "spinham.txt"

        output_name = os.path.join(out_head, out_tail)

        # Create the output directory if it does not exist
        if out_head != "":
            os.makedirs(out_head, exist_ok=True)

    # Read the model and the template
    model = load_tb2j_model(input_filename, quiet=not verbose)
    if template_file is not None:
        template = load_template(template_file)
    else:
        template = None

    model.filter(
        max_distance=max_distance,
        min_distance=min_distance,
        template=template,
    )

    additional_stats = (
        f"Exchange values are extracted from:\n"
        + f"{os.path.abspath(input_filename)}\n"
    )
    if template_file is not None:
        additional_stats += (
            f"Template used for "
            + ("filtering" if not form_model else "model formation")
            + ":\n"
            + f"{os.path.abspath(template_file)}\n"
        )
    # Remove template if no model formation is required
    if not form_model:
        template = None

    # Write the summary to the file or print it
    dump_spinham_txt(
        model,
        filename=output_name,
        anisotropic=not no_anisotropic,
        matrix=not no_matrix,
        dmi=not nodmi,
        template=template,
        decimals=decimals,
        additional_stats=additional_stats,
    )
    if output_name is not None:
        cprint(
            f"Extracted exchange info is in " + f"{os.path.abspath(output_name)}",
            "green",
        )


def create_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "-if",
        "--input-filename",
        required=True,
        type=str,
        help='Relative or absolute path to the "exchange.out" file,',
    )
    parser.add_argument(
        "-tf",
        "--template-file",
        default=None,
        type=str,
        help="Relative or absolute path to the template file,",
    )
    parser.add_argument(
        "-on",
        "--output-name",
        default=None,
        type=str,
        help="Name of the output files.",
    )
    parser.add_argument(
        "-d",
        "--decimals",
        default=4,
        type=int,
        help="Decimals after the comma for the exchange values.",
    )
    parser.add_argument(
        "-fm",
        "--form-model",
        default=False,
        action="store_true",
        help="Whether to form the model from the template.",
    )
    parser.add_argument(
        "-noa",
        "--no-anisotropic",
        default=False,
        action="store_true",
        help="Whether to output anisotropic exchange.",
    )
    parser.add_argument(
        "-nom",
        "--no-matrix",
        default=False,
        action="store_true",
        help="Whether to output whole matrix exchange.",
    )
    parser.add_argument(
        "-nodmi",
        default=False,
        action="store_true",
        help="Whether to output DMI exchange.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Verbose output, propagates to the called methods.",
    )
    parser.add_argument(
        "-maxd",
        "--max-distance",
        default=None,
        type=float,
        help="(<=) Maximum distance.",
    )
    parser.add_argument(
        "-mind",
        "--min-distance",
        default=None,
        type=float,
        help="(>=) Minimum distance.",
    )

    return parser



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Spin Hamiltonian module.

Write a tutorial with docstring here.
"""

__all__ = ["SpinHamiltonian", "ExchangeHamiltonian"]

from copy import deepcopy
from typing import Iterable, Tuple

import numpy as np

from radtools.constants import TORADIANS
from radtools.crystal.atom import Atom
from radtools.crystal.crystal import Crystal
from radtools.exceptions import NotationError
from radtools.spinham.constants import PREDEFINED_NOTATIONS
from radtools.spinham.parameter import ExchangeParameter
from radtools.spinham.template import ExchangeTemplate


class SpinHamiltonian(Crystal):
    r"""
    Spin Hamiltonian.

    By default the notation of the spin Hamiltonian is not defined
    and could be different in different context.
    However, in could always be checked via :py:attr:`.notation`.
    In user-specific cases it is the responsibility of the user to
    set the interpretation of the Hamiltonian`s notation.

    For the predefined notations see :py:meth:`.notation`.

    Child of the :py:class:`.Crystal` class.

    .. versionchanged:: 0.8.0 Renamed from SpinHamiltonian to SpinHamiltonian.

    Notes
    -----
    You can still use the old name, but it is deprecated and will be removed in the future.

    Parameters
    ----------
    crystal : :py:class:`.Crystal`, optional
        Crystal on which the spin Hamiltonian is build.
        By default it is cubic (:math:`a=1`) lattice with no atoms.
    notation : str or tuple of two bool and one float, optional
        One of the predefined notations or tuple with custom notation.
        See :py:attr:`.notation` for details.
    **kwargs
        Keyword arguments for the :py:class:`.Crystal` constructor.

    """

    def __init__(self, crystal: Crystal = None, notation=None, **kwargs) -> None:
        if crystal is not None:
            kwargs["atoms"] = crystal.atoms
            kwargs["cell"] = crystal.cell

        super().__init__(**kwargs)

        self._bonds = {}

        # Notation settings
        self._double_counting = None
        self._spin_normalized = None
        self._factor = None
        if notation is not None:
            self.notation = notation

    def __len__(self):
        return self._bonds.__len__()

    # Notation attributes
    @property
    def notation(self):
        r"""
        Tuple of the notation.

        It can be set with a

        * string
            One of the predefined notations: "standard", "TB2J", "SpinW"

        * iterable with 3 elements
            First two elements are converted to ``bool``,
            third element is interpreted as a float.
            Order: (double counting, spin normalized, factor).

        See :ref:`library_spinham_notation` for the detailed description.


        See Also
        --------
        double_counting
        spin_normalized
        factor
        set_interpretation
        """

        return (self.double_counting, self.spin_normalized, self.factor)

    @notation.setter
    def notation(self, new_notation):
        # Set the notation from predefined notations
        if isinstance(new_notation, str):
            new_notation = new_notation.lower()
            if new_notation not in PREDEFINED_NOTATIONS:
                raise ValueError(
                    f"Predefine notations are: "
                    + f"{list(PREDEFINED_NOTATIONS)}, got: {new_notation}"
                )
            new_notation = PREDEFINED_NOTATIONS[new_notation]
        # Set the notation from three values, converted to bool
        elif isinstance(new_notation, Iterable) and len(new_notation) == 3:
            new_notation = (
                bool(new_notation[0]),
                bool(new_notation[1]),
                float(new_notation[2]),
            )
        else:
            raise ValueError(
                "New notation has to be either a string "
                + "or an iterable with three elements, "
                + f"got: {new_notation}"
            )
        (self.double_counting, self.spin_normalized, self.factor) = new_notation

    @property
    def notation_string(self):
        r"""
        Human-readable representation of the notation.

        Returns
        -------
        notation_string : str
            Human-readable representation of the notation.
        """

        text_result = "H = "

        if self.factor != 1:
            if self.factor == -1:
                text_result += "- "
            elif self.factor == int(self.factor):
                text_result += f"{self.factor:.0f} "
            elif self.factor == 0.5:
                text_result += "1/2 "
            elif self.factor == -0.5:
                text_result += "-1/2 "
            else:
                text_result += f"{self.factor} "

        text_result += "\sum_{"
        if self.double_counting:
            text_result += "i,j}"
        else:
            text_result += "i>=j}"

        if self.spin_normalized:
            text_result += " e_i J_{ij} e_j\n"
        else:
            text_result += " S_i J_{ij} S_j\n"

        if self.double_counting:
            text_result += "Double counting is present.\n"
        else:
            text_result += "No double counting.\n"

        if self.spin_normalized:
            text_result += "Spin vectors are normalized to 1."
        else:
            text_result += "Spin vectors are not normalized."

        return text_result

    def set_interpretation(
        self, double_counting: bool = None, spin_normalized: bool = None, factor=None
    ):
        r"""
        Set the interpretation of the Hamiltonian`s notation.

        It is not changing the J values,
        but rather indicate how to interpret the J values of the Hamiltonian.

        Parameters
        ----------
        double_counting : bool, optional
        spin_normalized : bool, optional
        factor : float or int, optional

        See Also
        --------
        double_counting
        spin_normalized
        factor
        notation
        """

        if double_counting is not None:
            self._double_counting = bool(double_counting)
        if spin_normalized is not None:
            self._spin_normalized = bool(spin_normalized)
        if factor is not None:
            self._factor = factor

    @property
    def double_counting(self) -> bool:
        r"""
        Whether double counting is present in the Hamiltonian.

        Access this attribute to check the current notation of the Hamiltonian,
        set this attribute to change the notation of the Hamiltonian.

        Returns
        -------
        double_counting : bool
            ``True`` if double counting is present, ``False`` otherwise.

        Raises
        ------
        :py:exc:`.NotationError`
            If the interpretation of the Hamiltonian`s notation is not set.

        See Also
        --------
        set_interpretation : to change the interpretation of the Hamiltonian`s notation
            without the modification of the parameters.
        """

        if self._double_counting is None:
            raise NotationError("double_counting")
        return self._double_counting

    def _ensure_double_counting(self):
        r"""
        Ensure double counting.

        Check and fix if needed, that for each (atom1, atom2, R)
        (atom2, atom1, -R) is present in the Hamiltonian.
        """

        bonds = list(self)
        for atom1, atom2, (i, j, k), J in bonds:
            if (atom2, atom1, (-i, -j, -k)) not in self:
                self.add_bond(atom2, atom1, (-i, -j, -k), J=J.T)

    def _ensure_no_double_counting(self):
        r"""
        Ensure that there is no double counting in the model.

        If the bond is (atom1, atom2, (i,j,k)), then the bond is kept in the model
        if one of the conditions is true:

        * i > 0
        * i = 0 and j > 0
        * i = 0, j = 0 and k > 0
        * i = 0, j = 0, k = 0 and atom1.index <= atom2.index
        """

        bonds = list(self)

        for atom1, atom2, (i, j, k), J in bonds:
            if (i, j, k) != (0, 0, 0) or atom1 != atom2:
                if (
                    i > 0
                    or (i == 0 and j > 0)
                    or (i == 0 and j == 0 and k > 0)
                    or (i == 0 and j == 0 and k == 0 and atom1.index <= atom2.index)
                ):
                    if (atom2, atom1, (-i, -j, -k)) in self:
                        self.remove_bond(atom2, atom1, (-i, -j, -k))
                else:
                    if (atom2, atom1, (-i, -j, -k)) not in self:
                        self.add_bond(atom2, atom1, (-i, -j, -k), J=J.T)
                        self.remove_bond(atom1, atom2, (i, j, k))

    @double_counting.setter
    def double_counting(self, new_value: bool):
        # Need to be at the beginning, otherwise _ensure methods do not work
        old_value = self._double_counting
        self._double_counting = bool(new_value)

        if old_value is not None:
            factor = 1
            if old_value and not new_value:
                factor = 2
            elif not old_value and new_value:
                factor = 0.5
            if factor != 1:
                if new_value:
                    self._ensure_double_counting()
                else:
                    self._ensure_no_double_counting()
                for atom1, atom2, R, J in self:
                    if R != (0, 0, 0) or atom1 != atom2:
                        self[atom1, atom2, R] = J * factor
        else:
            if new_value:
                self._ensure_double_counting()
            else:
                self._ensure_no_double_counting()

    @property
    def spin_normalized(self) -> bool:
        r"""
        Whether spin is normalized.

        Access this attribute to check the current notation of the Hamiltonian,
        set this attribute to change the notation of the Hamiltonian.

        Returns
        -------
        spin_normalized : bool
            ``True`` if spin is normalized, ``False`` otherwise.

        Raises
        ------
        :py:exc:`.NotationError`
            If the interpretation of the Hamiltonian`s notation is not set.

        See Also
        --------
        set_interpretation : to change the interpretation of the Hamiltonian`s notation
            without the modification of the parameters.
        """

        if self._spin_normalized is None:
            raise NotationError("spin_normalized")
        return self._spin_normalized

    @spin_normalized.setter
    def spin_normalized(self, new_value: bool):
        if self._spin_normalized is not None:
            if self._spin_normalized and not new_value:
                for atom1, atom2, R, J in self:
                    self[atom1, atom2, R] = J / atom1.spin / atom2.spin
            elif not self._spin_normalized and new_value:
                for atom1, atom2, R, J in self:
                    self[atom1, atom2, R] = J * atom1.spin * atom2.spin
        self._spin_normalized = bool(new_value)

    @property
    def factor(self) -> float:
        r"""
        Whether any factor is present in the Hamiltonian.

        Access this attribute to check the current notation of the Hamiltonian,
        set this attribute to change the notation of the Hamiltonian.

        Returns
        -------
        factor : float
            value of the factor before the sum in the Hamiltonian.

        Raises
        ------
        :py:exc:`.NotationError`
            If the interpretation of the Hamiltonian`s notation is not set.

        See Also
        --------
        set_interpretation : to change the interpretation of the Hamiltonian`s notation
            without the modification of the parameters.
        """

        if self._factor is None:
            raise NotationError("factor")
        return self._factor

    @factor.setter
    def factor(self, new_factor: bool):
        if self._factor is not None:
            factor = self._factor / new_factor

            if factor != 1:
                for atom1, atom2, R, J in self:
                    self[atom1, atom2, R] = J * factor

        self._factor = float(new_factor)

    def __iter__(self):
        return SpinHamiltonianIterator(self)

    def __contains__(self, key):
        atom1, atom2, R = key
        # If atom is a string, get the atom object
        if isinstance(atom1, str):
            atom1 = self.get_atom(atom1)
        if isinstance(atom2, str):
            atom2 = self.get_atom(atom2)

        key = (atom1, atom2, R)
        return key in self._bonds

    def __getitem__(self, key) -> ExchangeParameter:
        atom1, atom2, R = key
        # If atom is a string, get the atom object
        if isinstance(atom1, str):
            atom1 = self.get_atom(atom1)
        if isinstance(atom2, str):
            atom2 = self.get_atom(atom2)

        key = (atom1, atom2, R)
        return self._bonds[key]

    def __getattr__(self, name):
        # Fix copy/deepcopy RecursionError
        if name in ["__setstate__"]:
            raise AttributeError(name)
        raise AttributeError(name)

    @property
    def crystal(self) -> Crystal:
        r"""
        Crystal of the Hamiltonian.

        Return an independent instance of a crystal.
        You can use it to play with the model`s crystal independently,
        but it will not affect the model itself.

        Returns
        -------
        crystal : :py:class:`.Crystal`
            Crystal of the Hamiltonian.

        See Also
        --------
        Crystal
        """
        return Crystal(self.lattice, self.atoms)

    @property
    def cell_list(self):
        r"""
        List of cells from the Hamiltonian.

        Return the list of R for all cell that are present in the Hamiltonian.
        Not ordered.

        Returns
        -------
        cells : (n, 3) :numpy:`ndarray`
            Array of n unit cells.
        """

        cells = set()
        for atom1, atom2, R, J in self:
            cells.add(R)
        return np.array(list(cells))

    @property
    def magnetic_atoms(self):
        r"""
        Magnetic atoms of the model.

        Atoms with at least one bond starting or finishing in it.

        Atoms are ordered with respect to the :py:attr:`.Atom.index`.

        Returns
        -------
        magnetic_atoms : list of :py:class:`.Atom`
            List of magnetic atoms.
        """
        result = set()
        for atom1, atom2, R, J in self:
            result.add(atom1)
            result.add(atom2)

        return sorted(list(result), key=lambda x: x.index)

    @property
    def number_spins_in_unit_cell(self):
        r"""
        Number of spins (or magnetic atoms) in the unit cell.

        Returns
        -------
        number_spins_in_unit_cell : int
            Number of spins (or magnetic atoms) in the unit cell.
        """

        return len(self.magnetic_atoms)

    @property
    def space_dimensions(self):
        r"""
        Model minimum and maximum coordinates in real space.

        Takes into account only an atom which has at least
        one bond associated with it.

        Returns
        -------
        x_min : float
            Minimum x coordinate.
        y_min : float
            Minimum y coordinate.
        z_min : float
            Minimum z coordinate.
        x_max : float
            Maximum x coordinate.
        y_max : float
            Maximum y coordinate.
        z_max : float
            Maximum z coordinate.
        """

        x_max, y_max, z_max = None, None, None
        x_min, y_min, z_min = None, None, None
        for atom1, atom2, R in self._bonds:
            x1, y1, z1 = self.get_atom_coordinates(atom1, relative=False)
            x2, y2, z2 = self.get_atom_coordinates(atom2, R, relative=False)
            if x_max is None:
                x_min, y_min, z_min = x1, y1, z1
                x_max, y_max, z_max = x1, y1, z1
            x_max = max(x1, x2, x_max)
            y_max = max(y1, y2, y_max)
            z_max = max(z1, z2, z_max)
            x_min = min(x1, x2, x_min)
            y_min = min(y1, y2, y_min)
            z_min = min(z1, z2, z_min)
        return x_min, y_min, z_min, x_max, y_max, z_max

    def __setitem__(self, key, value):
        self.add_bond(*key, value)

    def add_bond(
        self, atom1: Atom, atom2: Atom, R, J: ExchangeParameter = None, **kwargs
    ):
        r"""
        Add one bond to the Hamiltonian.

        More straightforward syntax is advised:

        .. doctest::

            >>> import radtools as rad
            >>> Cr = rad.Atom("Cr")
            >>> J = rad.ExchangeParameter(iso=1)
            >>> model = rad.SpinHamiltonian(rad.Crystal())
            >>> model[Cr, Cr, (1,0,0)] = J
            >>> (Cr, Cr, (1,0,0)) in model
            True

        It is equivalent to

        .. doctest::

            >>> import radtools as rad
            >>> Cr = rad.Atom("Cr")
            >>> J = rad.ExchangeParameter(iso=1)
            >>> model = rad.SpinHamiltonian(rad.Crystal())
            >>> model.add_bond(Cr, Cr, (1,0,0), J=J)
            >>> (Cr, Cr, (1,0,0)) in model
            True

        Parameters
        ----------
        atom1 : :py:class:`Atom` or str
            Atom object in (0, 0, 0) unit cell.
            str works only if atom is already in the Hamiltonian.
        atom2 : :py:class:`Atom` or str
            Atom object in R unit cell.
            str works only if atom is already in the Hamiltonian.
        R : tuple of ints
            Vector of the unit cell for atom2.
            In the relative coordinates (i,j,k).
        J : :py:class:`.ExchangeParameter`, optional
            An instance of :py:class:`ExchangeParameter`.
        ** kwargs
            Keyword arguments for the constructor of :py:class:`ExchangeParameter`.
            Ignored if J is given.
        """

        if isinstance(atom1, str):
            atom1 = self.get_atom(atom1)
        elif atom1 not in self.atoms:
            self.add_atom(atom1)

        if isinstance(atom2, str):
            atom2 = self.get_atom(atom2)
        elif atom2 not in self.atoms:
            self.add_atom(atom2)

        if J is None:
            J = ExchangeParameter(**kwargs)

        self._bonds[(atom1, atom2, R)] = J

        # Check for double counting
        double_counting = False
        try:
            double_counting = self.double_counting
        except NotationError:
            pass
        i, j, k = R
        if double_counting and (atom2, atom1, (-i, -j, -k)) not in self:
            self._bonds[(atom2, atom1, (-i, -j, -k))] = J.T

    def __delitem__(self, key):
        self.remove_bond(*key)

    def remove_bond(self, atom1: Atom, atom2: Atom, R):
        r"""
        Remove one bond from the Hamiltonian.

        More straightforward syntax is advised:

        .. doctest::

            >>> import radtools as rad
            >>> Cr = rad.Atom("Cr")
            >>> J = rad.ExchangeParameter(iso=1)
            >>> model = rad.SpinHamiltonian(rad.Crystal())
            >>> model[Cr, Cr, (1,0,0)] = J
            >>> (Cr, Cr, (1,0,0)) in model
            True
            >>> del model[Cr, Cr, (1,0,0)]
            >>> (Cr, Cr, (1,0,0)) in model
            False


        It is the same as

        .. doctest::

            >>> import radtools as rad
            >>> Cr = rad.Atom("Cr")
            >>> J = rad.ExchangeParameter(iso=1)
            >>> model = rad.SpinHamiltonian(rad.Crystal())
            >>> model[Cr, Cr, (1,0,0)] = J
            >>> (Cr, Cr, (1,0,0)) in model
            True
            >>> model.remove_bond(Cr, Cr, (1,0,0))
            >>> (Cr, Cr, (1,0,0)) in model
            False

        Parameters
        ----------
        atom1 : py:class:`.Atom`
            Atom object in (0, 0, 0) unit cell.
        atom2 : py:class:`.Atom`
            Atom object in R unit cell.
        R : tuple of ints
            Radius vector of the unit cell for atom2 (i,j,k).
        """

        # If atom is a string, get the atom object
        if isinstance(atom1, str):
            atom1 = self.get_atom(atom1)
        if isinstance(atom2, str):
            atom2 = self.get_atom(atom2)

        try:
            del self._bonds[(atom1, atom2, R)]
        except KeyError:
            raise KeyError(
                f"Bond ({atom2.fullname}, {atom2.fullname}, {R}) is not present in the model."
            )

        # Check for double counting
        double_counting = False
        try:
            double_counting = self.double_counting
        except NotationError:
            pass
        i, j, k = R
        if double_counting and (atom2, atom1, (-i, -j, -k)) in self:
            del self._bonds[(atom2, atom1, (-i, -j, -k))]

    def remove_atom(self, atom):
        r"""
        Remove magnetic atom from the Hamiltonian.

        Note: this method will remove atom itself and all the
        bonds, which starts or ends in this atom, if atom is magnetic.

        Parameters
        ----------
        atom : :py:class:`.Atom`
            Atom object.
        """

        # If atom given as a string, get the atom object
        if isinstance(atom, str):
            atom = self.get_atom(atom)

        bonds_for_removal = []
        for atom1, atom2, R, J in self:
            if atom1 == atom or atom2 == atom:
                bonds_for_removal.append((atom1, atom2, R))

        for bond in bonds_for_removal:
            del self[bond]

        super().remove_atom(atom)

    def filter(
        self, max_distance=None, min_distance=None, template=None, R_vector=None
    ):
        r"""
        Filter the exchange entries based on the given conditions.

        The result will be defined by logical conjugate of the conditions.
        Saying so the filtering will be performed for each given condition
        one by one.

        Parameters
        ----------
        max_distance : float or int, optional
            Distance for sorting, the condition is <<less or equal>>.
        min_distance : float or int, optional
            Distance for sorting, the condition is <<more or equal>>.
        template : list or :py:class:`.ExchangeTemplate`.
            List of pairs, which will remain in the Hamiltonian. ::

                [(atom1, atom2, R), ...]

        R_vector : tuple of ints or list of tuples of ints
            Tuple of 3 integers or list of tuples, specifying the R vectors,
            which will be kept after filtering.

        See Also
        --------
        filtered : Returns new object.

        Notes
        -----
        This method modifies the instance at which it is called.
        """

        if isinstance(template, ExchangeTemplate):
            template = template.get_list()
        if template is not None:
            template = set(template)

        if R_vector is not None:
            if type(R_vector) == tuple:
                R_vector = {R_vector}
            elif type(R_vector) == list:
                R_vector = set(R_vector)
            else:
                raise TypeError("Type is not supported, supported: list.")
        bonds_for_removal = set()
        for atom1, atom2, R, J in self:
            i, j, k = R
            dis = self.get_distance(atom1, atom2, R)

            if max_distance is not None and dis > max_distance:
                bonds_for_removal.add((atom1, atom2, R))

            if min_distance is not None and dis < min_distance:
                bonds_for_removal.add((atom1, atom2, R))

            # This behaviour should depend on the notation of the Hamiltonian
            condition = R_vector is not None and R not in R_vector
            try:
                if self.double_counting:
                    condition = condition and (-i, -j, -k) not in R_vector
            except NotationError:
                pass
            if condition:
                bonds_for_removal.add((atom1, atom2, R))

            # Here names, not objects are compared, because in general
            # template only has information about names (or fullnames) and R.
            if (
                template is not None
                and (atom1.name, atom2.name, R) not in template
                and (atom1.fullname, atom2.fullname, R) not in template
            ):
                bonds_for_removal.add((atom1, atom2, R))

        for atom1, atom2, R in bonds_for_removal:
            try:
                del self[atom1, atom2, R]
            except KeyError:
                pass

    def filtered(
        self, max_distance=None, min_distance=None, template=None, R_vector=None
    ):
        r"""
        Create filtered spin Hamiltonian based on the given conditions.

        The result will be defined by logical conjugate of the conditions.
        Saying so the filtering will be performed for each given condition
        one by one.
        Note: this method is not modifying the instance at which it is called.
        It will create a new instance with sorted :py:attr:`bonds` and all
        the other attributes will be copied (through :py:func:`deepcopy`).

        Parameters
        ----------
        max_distance : float or int, optional
            Distance for sorting, the condition is <<less or equal>>.
        min_distance : float or int, optional
            Distance for sorting, the condition is <<more or equal>>.
        template : list or :py:class:`.ExchangeTemplate`
            List of pairs, which will remain in the Hamiltonian. ::

                [(atom1, atom2, R), ...]

        R_vector : tuple of ints or list of tuples of ints
            Tuple of 3 integers or list of tuples, specifying the R vectors,
            which will be kept after filtering.

        Returns
        -------
        filtered_model : :py:class:`.SpinHamiltonian`
            Spin Hamiltonian after filtering.

        See Also
        --------
        filter : Modifies current object.

        Notes
        -----
        This method is not modifying the instance at which it is called.
        It creates a new instance (through deepcopy).
        """

        filtered_model = deepcopy(self)
        filtered_model.filter(
            max_distance=max_distance,
            min_distance=min_distance,
            template=template,
            R_vector=R_vector,
        )
        return filtered_model

    # TODO rewrite with new template logic (J1 through getattr)
    def form_model(self, template):
        r"""
        Force the Hamiltonian to have the symmetries of the template.

        Takes mean values of the parameters.
        For DMI interaction directions are kept the same,
        but the length of the DMI vector is scaled.

        This method modifies an instance on which it was called.

        .. versionchanged:: 0.8.0 Renamed from ``force_symmetry``

        Parameters
        ----------
        template : :py:class:`.ExchangeTemplate`
            Template.

        See Also
        --------
        formed_model : Returns new object.
        """

        if not isinstance(template, ExchangeTemplate):
            raise TypeError

        self.filter(template=template)

        for name in template.names:
            bonds = template.names[name]
            symm_matrix = np.zeros((3, 3), dtype=float)
            dmi_module = 0
            for atom1, atom2, R in bonds:
                # Here names, not objects are obtained, because in general
                # template only has information about names and R.
                J = self[self.get_atom(atom1), self.get_atom(atom2), R]
                symm_matrix = symm_matrix + J.symm_matrix
                dmi_module = dmi_module + J.dmi_module

            symm_matrix = symm_matrix / len(bonds)
            dmi_module = dmi_module / len(bonds)

            for atom1, atom2, R in bonds:
                J = self[self.get_atom(atom1), self.get_atom(atom2), R]
                if J.dmi_module != 0:
                    asymm_factor = dmi_module / J.dmi_module
                else:
                    asymm_factor = 0
                J.matrix = symm_matrix + J.asymm_matrix * asymm_factor

    def formed_model(self, template):
        r"""
        Form the model from the Hamiltoian based on the template.

        Takes mean values of the parameters.
        Respect the direction of the DMI vectors.

        This method return a new instance.

        Parameters
        ----------
        template : :py:class:`.ExchangeTemplate`
            Template.

        Returns
        -------
        new_model : :py:class:`.SpinHamiltonian`
            Spin Hamiltonian with forced symmetry.

        See Also
        --------
        form_model: Modifies current object.
        """

        new_model = deepcopy(self)
        new_model.form_model(template=template)
        return new_model

    def ferromagnetic_energy(self, theta=0, phi=0):
        r"""
        Compute energy of the Hamiltonian assuming ferromagnetic state.

        Notes
        -----
        Notation of the Hamiltonian has to be known.
        See :py:meth:`.set_interpretation` and :py:attr:`.notation`.

        Parameters
        ----------
        theta : float or (N,) |array_like|_, default 0
            Angle between z axis and direction of the magnetization.
            :math:`0^{\circ} \le \theta \le 180^{\circ}`.
        phi : float or (N,) |array_like|_, default 0
            angle between x axis and projection of
            magnetization`s direction on xy plane.
            :math:`0^{\circ} \le \phi < 360^{\circ}`.

        Returns
        -------
        energy : float or (N,) :numpy:`ndarray`
            Energy of ferromagnetic Hamiltonian with magnetization direction defined
            by ``theta`` and ``phi``. In the units of J values.
        """

        # Compute spin direction
        theta = np.array(theta) * TORADIANS
        phi = np.array(phi) * TORADIANS
        if theta.shape == ():
            theta = np.array([theta])
        if phi.shape == ():
            phi = np.array([phi])
        spin_direction = np.array(
            [np.cos(phi) * np.sin(theta), np.sin(phi) * np.sin(theta), np.cos(theta)]
        )

        energy = np.zeros((3, 3), dtype=float)
        for atom1, atom2, R, J in self:
            if self.spin_normalized:
                energy += self.factor * J.matrix
            else:
                energy += self.factor * J.matrix * atom1.spin * atom2.spin

        energy = np.einsum("ni,ij,jn->n", spin_direction.T, energy, spin_direction)
        if len(energy) == 1:
            energy = energy[0]
        return energy

    def input_for_magnons(self, nodmi=False, noaniso=False, custom_mask=None):
        r"""
        Input from the spin Hamiltonian.

        This function prepare the list of exchange parameters to
        be used as an input for magnon dispersion calculation.

        Parameters
        ----------
        nodmi : bool, default=False
            If True, then DMI is not included in the dispersion.
        noaniso : bool, default=False
            If True, then anisotropy is not included in the dispersion.
        custom_mask : func
            Custom mask for the exchange parameter. Function which take (3,3) numpy:`ndarray`
            as an input and returns (3,3) numpy:`ndarray` as an output.

        Returns
        -------
        Jij : list
        i : list
        j : list
        dij : list
        """

        Jij = []
        i = []
        j = []
        dij = []
        magnetic_atoms = self.magnetic_atoms
        atom_index = dict([(atom, i) for i, atom in enumerate(magnetic_atoms)])
        for atom1, atom2, R, J in self:
            if custom_mask is not None:
                result = custom_mask(J.matrix)
            else:
                result = J.matrix
                if nodmi:
                    result -= J.dmi_matrix
                if noaniso:
                    result -= J.aniso
            Jij.append(result)
            i.append(atom_index[atom1])
            j.append(atom_index[atom2])
            dij.append(self.get_vector(atom1, atom1, R))

        return Jij, i, j, dij


class ExchangeHamiltonian(SpinHamiltonian):
    pass


class SpinHamiltonianIterator:
    def __init__(self, exchange_model: SpinHamiltonian) -> None:
        self._bonds = list(
            map(
                lambda x: (x[0], x[1], x[2], exchange_model._bonds[x]),
                exchange_model._bonds,
            )
        )
        self._index = 0

    def __next__(self) -> Tuple[Atom, Atom, tuple, ExchangeParameter]:
        if self._index < len(self._bonds):
            result = self._bonds[self._index]
            self._index += 1
            return result
        raise StopIteration

    def __iter__(self):
        return self



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from copy import deepcopy
from math import sqrt

import numpy as np
from scipy.spatial.transform import Rotation

from radtools.crystal.kpoints import Kpoints
from radtools.geometry import span_orthonormal_set
from radtools.magnons.diagonalization import ColpaFailed, solve_via_colpa
from radtools.spinham.hamiltonian import SpinHamiltonian

__all__ = ["MagnonDispersion"]


class MagnonDispersion:
    r"""
    Magnon dispersion wrapper.

    Parameters
    ----------
    model : :py:class:`.SpinHamiltonian`
        Spin Hamiltonian.
    Q : (3,) |array_like|_
        Ordering wave vector of the spin-spiral.
        In relative coordinates with respect to the model`s reciprocal cell.
        It rotates spins from unit cell to unit cell,
        but not from atom to atom in (0, 0, 0) unit cell.
    n : (3,) |array_like|_, optional
        Global rotational axis. If None provided, then it is set to the direction of ``Q``.
    nodmi : bool, default=False
        If True, then DMI is not included in the dispersion.
    noaniso : bool, default=False
        If True, then anisotropy is not included in the dispersion.
    custom_mask : func
        Custom mask for the exchange parameter. Function which take (3,3) :numpy:`ndarray`
        as an input and returns (3,3) :numpy:`ndarray` as an output.

    Attributes
    ----------
    Q : (3,) :numpy:`ndarray`
        Ordering wave vector of the spin-spiral. in absolute coordinates in reciprocal space.
    n : (3,) :numpy:`ndarray`
        Global rotational axis.
    N : int
        Number of magnetic atoms.
    J_matrices : (M,) :numpy:`ndarray`
        Exchange parameters.
    i : (M,) :numpy:`ndarray`
        Indices of the first atom in the exchange pair.
    j : (M,) :numpy:`ndarray`
        Indices of the second atom in the exchange pair.
    d : (M, 3) :numpy:`ndarray`
        Vectors from the first atom to the second atom in the exchange pair.
    S : (N, 3) :numpy:`ndarray`
        Spin vectors.
    u : (N, 3) :numpy:`ndarray`
        Defined from local spin directions.
    v : (N, 3) :numpy:`ndarray`
        Defined from local spin directions.
    """

    def __init__(
        self,
        model: SpinHamiltonian,
        Q=None,
        n=None,
        nodmi=False,
        noaniso=False,
        custom_mask=None,
    ):
        self._C = None
        # Store the exchange model, but privately
        self._model = deepcopy(model)
        self._model.notation = "SpinW"

        # Convert Q to absolute coordinates
        if Q is None:
            Q = [0, 0, 0]
        self.Q = np.array(Q, dtype=float) @ self._model.reciprocal_cell

        # Convert n to absolute coordinates, use Q if n is not provided
        if n is None:
            if np.allclose([0, 0, 0], Q):
                self.n = np.array([0, 0, 1])
            else:
                self.n = self.Q / np.linalg.norm(self.Q)
        else:
            self.n = np.array(n, dtype=float) / np.linalg.norm(n)

        # Get the number of magnetic atoms
        self.N = len(self._model.magnetic_atoms)

        # Get the exchange parameters, indices and vectors form the SpinHamiltonian
        (
            self.J_matrices,
            self.indices_i,
            self.indices_j,
            self.dis_vectors,
        ) = self._model.input_for_magnons(
            nodmi=nodmi, noaniso=noaniso, custom_mask=custom_mask
        )

        # Initialize spin vector, u vector and v vector arrays
        self.S = np.zeros((self.N, 3), dtype=float)
        self.u = np.zeros((self.N, 3), dtype=complex)
        self.v = np.zeros((self.N, 3), dtype=complex)

        # Get spin vectors and compute u and v vectors from local spin directions
        for a_i, atom in enumerate(self._model.magnetic_atoms):
            try:
                self.S[a_i] = atom.spin_vector
                e1, e2, e3 = span_orthonormal_set(self.S[a_i])
                self.v[a_i] = e3
                self.u[a_i] = e1 + 1j * e2
            except ValueError:
                raise ValueError(
                    f"Spin vector is not defined for {atom.fullname} atom."
                )

        # Rotate exchange matrices
        for i in range(len(self.J_matrices)):
            rotvec = self.n * (self.Q @ self.dis_vectors[i])
            R_nm = Rotation.from_rotvec(rotvec).as_matrix()
            self.J_matrices[i] = self.J_matrices[i] @ R_nm

    def J(self, k):
        r"""
        Computes J(k) matrix.

        .. math::

            \boldsymbol{J}_{i,j}(\boldsymbol{k}) = \sum_{\boldsymbol{d}}\boldsymbol{J}_{i,j}(\boldsymbol{d})e^{-i\boldsymbol{k}\boldsymbol{d}}

        where indices :math:`i` and :math:`j` correspond to the atoms in the exchange pair.

        Parameters
        ----------
        k : (3,) |array_like|_
            Reciprocal vector.
            In absolute coordinates.
        """
        # Initialize matrix
        result = np.zeros((self.N, self.N, 3, 3), dtype=complex)
        k = np.array(k)
        # Compute J(k)
        for index in range(len(self.J_matrices)):
            i = self.indices_i[index]
            j = self.indices_j[index]

            result[i][j] += self.J_matrices[index] * np.exp(
                -1j * (k @ self.dis_vectors[index])
            )
        return result

    def A(self, k):
        r"""
        Computes A(k) matrix.

        .. math::

            A(\boldsymbol{k})^{ij} = \dfrac{\sqrt{S_i\cdot S_j}}{2}\boldsymbol{u}^T_i\boldsymbol{J}_{i,j}(-\boldsymbol{k})\overline{\boldsymbol{u}}_j

        where indices :math:`i` and :math:`j` correspond to the atoms in the exchange pair.

        Parameters
        ----------
        k : (3,) |array_like|_
            Reciprocal vector.
            In absolute coordinates.
        """
        # Initialize matrix
        result = np.zeros((self.N, self.N), dtype=complex)
        k = np.array(k)

        # Compute A(k)
        J = self.J(-k)
        for i in range(len(J)):
            for j in range(len(J[i])):
                result[i][j] += (
                    sqrt(np.linalg.norm(self.S[i]) * np.linalg.norm(self.S[j]))
                    / 2
                    * (self.u[i] @ J[i][j] @ np.conjugate(self.u[j]))
                )
        return result

    def B(self, k):
        r"""
        Computes B(k) matrix.

        .. math::

            B(\boldsymbol{k})^{ij} = \dfrac{\sqrt{S_i\cdot S_j}}{2}\boldsymbol{u}^T_i\boldsymbol{J}_{i,j}(-\boldsymbol{k})\boldsymbol{u}_j

        where indices :math:`i` and :math:`j` correspond to the atoms in the exchange pair.

        Parameters
        ----------
        k : (3,) |array_like|_
            Reciprocal vector.
            In absolute coordinates.
        """
        # Initialize matrix
        result = np.zeros((self.N, self.N), dtype=complex)
        k = np.array(k)

        # Compute B(k)
        J = self.J(-k)
        for i in range(len(J)):
            for j in range(len(J[i])):
                result[i][j] += (
                    sqrt(np.linalg.norm(self.S[i]) * np.linalg.norm(self.S[j]))
                    / 2
                    * (self.u[i] @ J[i][j] @ self.u[j])
                )
        return result

    def C(self):
        r"""
        Computes C matrix.

        .. math::

            C^{i,j} = \delta_{i,j}\sum_{l}S_l \boldsymbol{v}^T_i\boldsymbol{J}_{i, l}(\boldsymbol{0})\boldsymbol{v}_l

        where indices :math:`i` and :math:`j` correspond to the atoms in the exchange pair.
        """

        if self._C is None:
            self._C = np.zeros((self.N, self.N), dtype=complex)

            # Compute C matrix, note: sum over l is hidden here
            J = self.J(np.zeros(3))
            for i in range(len(J)):
                for j in range(len(J[i])):
                    self._C[i][i] += (
                        np.linalg.norm(self.S[j]) * self.v[i] @ J[i][j] @ self.v[j]
                    )
        return self._C

    def h(self, k):
        k = np.array(k)
        # Compute h matrix
        left = np.concatenate(
            (2 * self.A(k) - 2 * self.C(), 2 * np.conjugate(self.B(k)).T), axis=0
        )
        right = np.concatenate(
            (2 * self.B(k), 2 * np.conjugate(self.A(-k)) - 2 * self.C()), axis=0
        )
        h = np.concatenate((left, right), axis=1)
        return h

    def omega(self, k, zeros_to_none=False, return_G=False, return_imaginary=False):
        r"""
        Computes magnon energies.

        Parameters
        ----------
        k : (3,) |array_like|_
            Reciprocal vector.
            In absolute coordinates.
        zeros_to_none : bool, default=False
            If True, then return ``None`` instead of 0 if Colpa fails.
        return_G : bool, default False
            Whether to return the transformation matrix.
            See :py:func:`.solve_via_colpa` for details.
            .. versionadded:: 0.8.10
        return_imaginary : bool, default False
            Whether to return imaginary part of the energies.
            If ``True``, then real and imaginary part can be accessed via
            ``omegas.real`` and ``omegas.imag``.
            .. versionadded:: 0.8.10

        Returns
        -------
        omegas : (N,) :numpy:`ndarray`
            Magnon energies for the vector ``k``.
            If ``return_G`` is ``True``, then return (2N,) :numpy:`ndarray` else
            return (N,) :numpy:`ndarray`. See :py:func:`.solve_via_colpa` for details.
        G : (2N, 2N) : :numpy:`ndarray`
            Transformation matrix. Returned only if ``return_G`` is ``True``.
            See :py:func:`.solve_via_colpa` for details.
        """
        # Diagonalize h matrix via Colpa
        k = np.array(k)
        h = self.h(k)
        try:
            omegas, G = solve_via_colpa(h)
        except ColpaFailed:
            # Try to solve for positive semidefinite matrix
            try:
                omegas, G = solve_via_colpa(h + np.diag(1e-8 * np.ones(2 * self.N)))
            except ColpaFailed:
                # Try to solve for negative defined matrix
                try:
                    omegas, G = solve_via_colpa(-h)
                    omegas *= -1
                    G *= -1
                except ColpaFailed:
                    # Try to solve for negative semidefinite matrix
                    try:
                        omegas, G = solve_via_colpa(
                            -h - np.diag(1e-8 * np.ones(h.shape[0]))
                        )
                        omegas *= -1
                        G *= -1
                    except ColpaFailed:
                        # If all fails, return None or 0
                        if zeros_to_none:
                            omegas = np.array([None] * 2 * self.N, dtype=float)
                            if return_G:
                                G = np.empty((2 * self.N, 2 * self.N), dtype=float)

                        else:
                            omegas = np.zeros(self.N, dtype=float)
                            if return_G:
                                G = np.zeros((2 * self.N, 2 * self.N), dtype=float)
        omegas[np.abs(omegas) <= 1e-8] = 0
        if not return_imaginary:
            omegas = omegas.real
        if return_G:
            return omegas, G
        else:
            return omegas[: self.N]

    def omegas(self, kpoints, zeros_to_none=False):
        r"""
        Dispersion spectra.

        Parameters
        ----------
        kpoints : (M, 3) |array_like|_
            K points in absolute coordinates.
        zeros_to_none : bool, default=False
            If True, then return ``None`` instead of 0 if Colpa fails.
        """

        data = []

        if isinstance(kpoints, Kpoints):
            kpoints = kpoints.points()

        for point in kpoints:
            data.append(self.omega(point, zeros_to_none=zeros_to_none))

        return np.array(data).T

    def __call__(self, *args, **kwargs):
        return self.omegas(*args, **kwargs)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
RAD-tools
"""


__version__ = "0.8.13"
__doclink__ = "rad-tools.org"
__git_hash__ = "43a49a2aaa1472db5dd040fa5186cc293e9b7e45"
__release_date__ = "4 March 2024"


from . import (
    constants,
    crystal,
    decorate,
    dos,
    exceptions,
    geometry,
    io,
    magnons,
    numerical,
    score,
    spinham,
)
from .constants import *
from .crystal import *
from .decorate import *
from .dos import *
from .geometry import *
from .io import *
from .magnons import *
from .numerical import *
from .score import *
from .spinham import *

__all__ = ["__version__", "__doclink__", "__git_hash__", "__release_date__"]
__all__.extend(crystal.__all__)
__all__.extend(decorate.__all__)
__all__.extend(dos.__all__)
__all__.extend(spinham.__all__)
__all__.extend(io.__all__)
__all__.extend(magnons.__all__)
__all__.extend(score.__all__)
__all__.extend(constants.__all__)
__all__.extend(geometry.__all__)
__all__.extend(numerical.__all__)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from calendar import month_name
from datetime import datetime

from radtools import __doclink__, __git_hash__, __release_date__, __version__
from radtools._license import LICENSE

__all__ = ["logo", "stamp_line", "license"]


def logo(info=None, line_length=71, flat=False, date_time=False, comment=None):
    """
    Logo generator for rad-tools package.

    Returns the logo and information about the package.

    Parameters
    ----------
    info : list of str, optional
        Information about the package.
        will be displayed below the logo.
        Each element should not exceed 59 characters.
        by default it displays the version, release date,
        git hash and documentation link. You can pass th empty list
        to display the logo only.
    line_length : int
        Length of the lines to be returned.
        Minimum value is 71.
    flat : bool
        Whether to return a flat logo or not.
    date_time : bool, default False
        Whether to include the date and time to the standard info or not.
    comment : str or bool, optional
        Whether to use some character at the end of each string. If bool and
        True, then "# " is used. If str, then this string is used. If None, then
        no character is used.
    Returns
    -------
    logo_info : str
        Logo and information about the package.
    """
    if info is None:
        info = [
            f"Version: {__version__}",
            f"Documentation: {__doclink__}",
            f"Release date: {__release_date__}",
            f"Git hash: {__git_hash__}",
            f"Licence: GNU GPLv3",
        ]
        if date_time:
            cd = datetime.now()
            info.append("")
            info.append(
                f"Generated on {cd.day} {month_name[cd.month]} {cd.year}"
                + f" at {cd.hour}:{cd.minute}:{cd.second} "
            )
    logo = [
        "██████╗  █████╗ ██████╗       ████████╗ █████╗  █████╗ ██╗      ██████╗",
        "██╔══██╗██╔══██╗██╔══██╗      ╚══██╔══╝██╔══██╗██╔══██╗██║     ██╔════╝",
        "██████╔╝███████║██║  ██║█████╗   ██║   ██║  ██║██║  ██║██║     ╚═█████╗",
        "██╔══██╗██╔══██║██║  ██║╚════╝   ██║   ██║  ██║██║  ██║██║       ╚══██║",
        "██║  ██║██║  ██║██████╔╝         ██║   ╚█████╔╝╚█████╔╝███████╗██████╔╝",
        "╚═╝  ╚═╝╚═╝  ╚═╝╚═════╝          ╚═╝    ╚════╝  ╚════╝ ╚══════╝╚═════╝ ",
    ]
    if flat:
        logo = [
            "██████   █████  ██████        ████████  █████   █████  ██       ██████ ",
            "██   ██ ██   ██ ██   ██          ██    ██   ██ ██   ██ ██      ██      ",
            "██████  ███████ ██   ██ █████    ██    ██   ██ ██   ██ ██        █████ ",
            "██   ██ ██   ██ ██   ██          ██    ██   ██ ██   ██ ██           ██ ",
            "██   ██ ██   ██ ██████           ██     █████   █████  ███████ ██████  ",
        ]
    cat = [
        "▄   ▄     ",
        "█▀█▀█     ",
        "█▄█▄█     ",
        " ███   ▄▄ ",
        " ████ █  █",
        " ████    █",
        " ▀▀▀▀▀▀▀▀ ",
    ]

    N = 71
    n = 12
    if line_length < N:
        line_length = N
    if isinstance(comment, bool) and comment:
        comment = "# "
    elif comment is not None:
        comment = str(comment)
    else:
        comment = ""

    logo_info = [f"{x:^{N}}" for x in logo]
    if len(info) > 0:
        if len(info) <= len(cat):
            before = (len(cat) - len(info)) // 2 + (len(cat) - len(info)) % 2
            after = len(cat) - len(info) - before
            for i in range(len(cat)):
                if i < before or i >= len(cat) - after:
                    logo_info.append(f"{' ':{N-n}}{cat[i]:^{n}}")
                else:
                    logo_info.append(f"{info[i-before]:^{N-n}}{cat[i]:^{n}}")
        else:
            before = (len(info) - len(cat)) // 2
            after = len(info) - len(cat) - before
            for i in range(len(info)):
                if i < before or i >= len(info) - after:
                    logo_info.append(f"{info[i-before]:^{N-n}}")
                else:
                    logo_info.append(f"{info[i-before]:^{N-n}}{cat[i-before]:^{n}}")

    logo_info = [f"{comment}{x:^{line_length}}\n" for x in logo_info]
    return "".join(logo_info)[:-1]


def stamp_line(date_time=True, version=True, githash=False, doclink=False):
    """
    Return one-line information about the package.

    Parameters
    ----------
    date_time : bool, default True
        Whether to include the release date or not.
    version : bool, default True
        Whether to include the version number or not.
    githash : bool, default False
        Whether to include the git hash or not.
    doclink : bool, default False
        Whether to include the documentation link or not.

    Returns
    -------
    info : str
        Information about the package.
    """

    line = []
    if date_time:
        cd = datetime.now()
        line.append(
            f"on {cd.day} {month_name[cd.month]} {cd.year}"
            + f" at {cd.hour}:{cd.minute}:{cd.second} "
        )
    line.append("by rad-tools ")
    if version:
        line.append(f"{__version__} ")
    if githash:
        line.append(f"(githash {__git_hash__}) ")
    if doclink:
        line.append(f"Documentation: {__doclink__}")
    return "".join(line)


def license():
    """
    Return the license of the package.

    Returns
    -------
    license : str
        License of the package.
    """

    return LICENSE



#! /usr/local/bin/python3

from radtools._osfix import _winwait
from radtools.score.extract_tb2j import create_parser, manager

if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    manager(**vars(args))
    _winwait()



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from numpy.linalg import LinAlgError

from radtools.exceptions import ColpaFailed

__all__ = ["solve_via_colpa"]


def solve_via_colpa(D):
    r"""
    Diagonalize grand-dynamical matrix following the method of Colpa [1]_.

    Algorithm itself is described in section 3, Remark 1 of [1]_.

    Solves the Bogoliubov Hamiltonian of the form:

    .. math::

        \hat{H} = \sum_{r^{\prime}, r = 1}^m
        \hat{\alpha}_{r^{\prime}}^{\dagger}\boldsymbol{\Delta}_1^{r^{\prime}r}\hat{\alpha}_r +
        \hat{\alpha}_{r^{\prime}}^{\dagger}\boldsymbol{\Delta}_2^{r^{\prime}r}\hat{\alpha}_{m+r}^{\dagger} +
        \hat{\alpha}_{m+r^{\prime}}^{\dagger}\boldsymbol{\Delta}_3^{r^{\prime}r}\hat{\alpha}_r +
        \hat{\alpha}_{m+r^{\prime}}^{\dagger}\boldsymbol{\Delta}_4^{r^{\prime}r}\hat{\alpha}_{m+r}^{\dagger}

    In a matrix form the Hamiltonian is:

    .. math::

        \hat{H} = \boldsymbol{\hat{a}}^{\dagger} \boldsymbol{D} \boldsymbol{\hat{a}}

    where

    .. math::

        \boldsymbol{\hat{a}} =
        \begin{pmatrix}
            \hat{\alpha}_1 \\
            \cdots \\
            \hat{\alpha}_m \\
            \hat{\alpha}_{m+1} \\
            \cdots \\
            \hat{\alpha}_{2m} \\
        \end{pmatrix}

    After diagonalization the Hamiltonian is:

    .. math::

        \hat{H} = \boldsymbol{\hat{c}}^{\dagger} \boldsymbol{E} \boldsymbol{\hat{c}}

    Parameters
    ----------
    D : (2N, 2N) |array_like|_
        Grand dynamical matrix. Must be Hermitian and positive-defined.

        .. math::

            \boldsymbol{D} = \begin{pmatrix}
                \boldsymbol{\Delta_1} & \boldsymbol{\Delta_2} \\
                \boldsymbol{\Delta_3} & \boldsymbol{\Delta_4}
            \end{pmatrix}

    Returns
    -------
    E : (2N,) :numpy:`ndarray`
        The eigenvalues, each repeated according to its multiplicity.
        First N eigenvalues are sorted in descending order.
        Last N eigenvalues are sorted in ascending order.
        In the case of diagonalization of the magnon Hamiltonian
        first N eigenvalues are the same as last N eigenvalues, but
        in reversed order. It is an array of the diagonal elements of the
        diagonal matrix :math:`\boldsymbol{E}` from the diagonalized Hamiltonian.

    G : (2N, 2N) :numpy:`ndarray`
        Transformation matrix, which change the basis from the original set of bosonic
        operators :math:`\boldsymbol{\hat{a}}` to the set of
        new bosonic operators :math:`\boldsymbol{\hat{c}}` which diagonalize
        the Hamiltonian:

        .. math::

            \boldsymbol{\hat{c}} = \boldsymbol{G} \boldsymbol{\hat{a}}

    Notes
    -----

    Let :math:`\boldsymbol{E}` be the diagonal matrix of eigenvalues ``E``, then:

    .. math::

        \boldsymbol{E} = (\boldsymbol{G}^{\dagger})^{-1} \boldsymbol{D} \boldsymbol{G}^{-1}



    References
    ----------
    .. [1] Colpa, J.H.P., 1978.
        Diagonalization of the quadratic boson hamiltonian.
        Physica A: Statistical Mechanics and its Applications,
        93(3-4), pp.327-353.
    """

    D = np.array(D)

    N = len(D) // 2
    g = np.diag(np.concatenate((np.ones(N), -np.ones(N))))

    try:
        # In Colpa article decomposition is K^{\dag}K, while numpy gives KK^{\dag}
        K = np.conjugate(np.linalg.cholesky(D)).T
    except LinAlgError:
        raise ColpaFailed

    L, U = np.linalg.eig(K @ g @ np.conjugate(K).T)

    # Sort with respect to L, in descending order
    U = np.concatenate((L[:, None], U.T), axis=1).T
    U = U[:, np.argsort(U[0])]
    L = U[0, ::-1]
    U = U[1:, ::-1]

    E = g @ L

    G_minus_one = np.linalg.inv(K) @ U @ np.sqrt(np.diag(E))

    # Compute G from G^-1 following Colpa, see equation (3.7) for details
    G = np.conjugate(G_minus_one).T
    G[:N, N:] *= -1
    G[N:, :N] *= -1

    return E, G



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Crystal/lattice identification.
"""

from math import acos, cos, floor, log10, sqrt

import numpy as np
from termcolor import cprint

import radtools.crystal.cell as Cell
from radtools.constants import TODEGREES, TORADIANS
from radtools.decorate.array import print_2d_array
from radtools.geometry import volume
from radtools.numerical import compare_numerically

__all__ = ["niggli", "lepage"]


def niggli(
    a=1,
    b=1,
    c=1,
    alpha=90,
    beta=90,
    gamma=90,
    eps_rel=1e-5,
    verbose=False,
    return_cell=False,
    max_iter=10000,
):
    r"""
    Computes Niggli matrix form.

    Parameters
    ----------
    a : float, default 1
        Length of the :math:`a_1` vector.
    b : float, default 1
        Length of the :math:`a_2` vector.
    c : float, default 1
        Length of the :math:`a_3` vector.
    alpha : float, default 90
        Angle between vectors :math:`a_2` and :math:`a_3`. In degrees.
    beta : float, default 90
        Angle between vectors :math:`a_1` and :math:`a_3`. In degrees.
    gamma : float, default 90
        Angle between vectors :math:`a_1` and :math:`a_2`. In degrees.
    eps_rel : float, default 1e-5
        Relative epsilon as defined in [2]_.
    verbose : bool, default False
        Whether to print the steps of an algorithm.
    return_cell : bool, default False
        Whether to return cell parameters instead of Niggli matrix form.
    max_iter : int, default 100000
        Maximum number of iterations.

    Returns
    -------
    result : (3,2) :numpy:`ndarray`
        Niggli matrix form as defined in [1]_:

        .. math::

            \begin{pmatrix}
                A & B & C \\
                \xi/2 & \eta/2 & \zeta/2
            \end{pmatrix}

        If return_cell == True, then return Niggli cell: (a, b, c, alpha, beta, gamma).

    Raises
    ------
    ValueError
        If the niggli cell is not found in ``max_iter`` iterations.
    ValueError
        If the provided cell`s volume is zero.

    References
    ----------
    .. [1] Křivý, I. and Gruber, B., 1976.
        A unified algorithm for determining the reduced (Niggli) cell.
        Acta Crystallographica Section A: Crystal Physics, Diffraction,
        Theoretical and General Crystallography,
        32(2), pp.297-298.
    .. [2] Grosse-Kunstleve, R.W., Sauter, N.K. and Adams, P.D., 2004.
        Numerically stable algorithms for the computation of reduced unit cells.
        Acta Crystallographica Section A: Foundations of Crystallography,
        60(1), pp.1-6.

    Examples
    --------
    Example from [1]_
    (parameters are reproducing :math:`A=9`, :math:`B=27`, :math:`C=4`,
    :math:`\xi` = -5, :math:`\eta` = -4, :math:`\zeta = -22`):

    .. doctest::

        >>> import radtools as rad
        >>> from radtools.constants import TODEGREES
        >>> from math import acos, sqrt
        >>> a = 3
        >>> b = sqrt(27)
        >>> c = 2
        >>> print(f"{a} {b:.3f} {c}")
        3 5.196 2
        >>> alpha = acos(-5 / 2 / b / c) * TODEGREES
        >>> beta = acos(-4 / 2 / a / c) * TODEGREES
        >>> gamma = acos(-22 / 2 / a / b) * TODEGREES
        >>> print(f"{alpha:.2f} {beta:.2f} {gamma:.2f}")
        103.92 109.47 134.88
        >>> niggli_matrix_form = rad.niggli(a, b, c, alpha, beta, gamma, verbose=True)
                       A         B         C        xi        eta      zeta
        start:       9.00000  27.00000   4.00000  -5.00000  -4.00000 -22.00000
        2 appl. to   9.00000  27.00000   4.00000  -5.00000  -4.00000 -22.00000
        1 appl. to   9.00000   4.00000  27.00000  -5.00000 -22.00000  -4.00000
        4 appl. to   4.00000   9.00000  27.00000 -22.00000  -5.00000  -4.00000
        5 appl. to   4.00000   9.00000  27.00000 -22.00000  -5.00000  -4.00000
        4 appl. to   4.00000   9.00000  14.00000  -4.00000  -9.00000  -4.00000
        6 appl. to   4.00000   9.00000  14.00000  -4.00000  -9.00000  -4.00000
        4 appl. to   4.00000   9.00000   9.00000  -8.00000  -1.00000  -4.00000
        7 appl. to   4.00000   9.00000   9.00000  -8.00000  -1.00000  -4.00000
        3 appl. to   4.00000   9.00000   9.00000  -9.00000  -1.00000   4.00000
        5 appl. to   4.00000   9.00000   9.00000   9.00000   1.00000   4.00000
        3 appl. to   4.00000   9.00000   9.00000  -9.00000  -3.00000   4.00000
        result:      4.00000   9.00000   9.00000   9.00000   3.00000   4.00000
        >>> niggli_matrix_form
        array([[4. , 9. , 9. ],
               [4.5, 1.5, 2. ]])

    """
    cell_volume = volume(a, b, c, alpha, beta, gamma)
    if cell_volume == 0:
        raise ValueError("Cell volume is zero")
    eps = eps_rel * volume(a, b, c, alpha, beta, gamma) ** (1 / 3.0)
    n = abs(floor(log10(abs(eps))))

    # 0
    A = a**2
    B = b**2
    C = c**2
    xi = 2 * b * c * cos(alpha * TORADIANS)
    eta = 2 * a * c * cos(beta * TORADIANS)
    zeta = 2 * a * b * cos(gamma * TORADIANS)
    N = (
        max(
            len(str(A).split(".")[0]),
            len(str(B).split(".")[0]),
            len(str(C).split(".")[0]),
            len(str(xi).split(".")[0]),
            len(str(eta).split(".")[0]),
            len(str(zeta).split(".")[0]),
        )
        + 1
        + n
    )

    def summary_line():
        return (
            f"{A:{N}.{n}f} {B:{N}.{n}f} {C:{N}.{n}f} "
            + f"{xi:{N}.{n}f} {eta:{N}.{n}f} {zeta:{N}.{n}f}"
        )

    if verbose:
        print(
            f"           {'A':^{N}} {'B':^{N}} {'C':^{N}} "
            + f"{'xi':^{N}} {'eta':^{N}} {'zeta':^{N}}"
        )
        cprint(
            f"start:     {summary_line()}",
            color="yellow",
        )
        phrase = "appl. to"
    iter_count = 0
    while True:
        if iter_count > max_iter:
            raise ValueError(f"Niggli cell not found in {max_iter} iterations")
        iter_count += 1
        # 1
        if compare_numerically(A, ">", B, eps) or (
            compare_numerically(A, "==", B, eps)
            and compare_numerically(abs(xi), ">", abs(eta), eps)
        ):
            if verbose:
                print(f"1 {phrase} {summary_line()}")
            A, xi, B, eta = B, eta, A, xi
        # 2
        if compare_numerically(B, ">", C, eps) or (
            compare_numerically(B, "==", C, eps)
            and compare_numerically(abs(eta), ">", abs(zeta), eps)
        ):
            if verbose:
                print(f"2 {phrase} {summary_line()}")
            B, eta, C, zeta = C, zeta, B, eta
            # go to 1
            continue
        # 3
        if compare_numerically(xi * eta * zeta, ">", 0, eps):
            if verbose:
                print(f"3 {phrase} {summary_line()}")
            xi, eta, zeta = abs(xi), abs(eta), abs(zeta)
        # 4
        if compare_numerically(xi * eta * zeta, "<=", 0, eps):
            if verbose:
                print(f"4 {phrase} {summary_line()}")
            xi, eta, zeta = -abs(xi), -abs(eta), -abs(zeta)
        # 5
        if (
            compare_numerically(abs(xi), ">", B, eps)
            or (
                compare_numerically(xi, "==", B, eps)
                and compare_numerically(2 * eta, "<", zeta, eps)
            )
            or (
                compare_numerically(xi, "==", -B, eps)
                and compare_numerically(zeta, "<", 0, eps)
            )
        ):
            if verbose:
                print(f"5 {phrase} {summary_line()}")
            C = B + C - xi * np.sign(xi)
            eta = eta - zeta * np.sign(xi)
            xi = xi - 2 * B * np.sign(xi)
            # go to 1
            continue
        # 6
        if (
            compare_numerically(abs(eta), ">", A, eps)
            or (
                compare_numerically(eta, "==", A, eps)
                and compare_numerically(2 * xi, "<", zeta, eps)
            )
            or (
                compare_numerically(eta, "==", -A, eps)
                and compare_numerically(zeta, "<", 0, eps)
            )
        ):
            if verbose:
                print(f"6 {phrase} {summary_line()}")
            C = A + C - eta * np.sign(eta)
            xi = xi - zeta * np.sign(eta)
            eta = eta - 2 * A * np.sign(eta)
            # go to 1
            continue
        # 7
        if (
            compare_numerically(abs(zeta), ">", A, eps)
            or (
                compare_numerically(zeta, "==", A, eps)
                and compare_numerically(2 * xi, "<", eta, eps)
            )
            or (
                compare_numerically(zeta, "==", -A, eps)
                and compare_numerically(eta, "<", 0, eps)
            )
        ):
            if verbose:
                print(f"7 {phrase} {summary_line()}")
            B = A + B - zeta * np.sign(zeta)
            xi = xi - eta * np.sign(zeta)
            zeta = zeta - 2 * A * np.sign(zeta)
            # go to 1
            continue
        # 8
        if compare_numerically(xi + eta + zeta + A + B, "<", 0, eps) or (
            compare_numerically(xi + eta + zeta + A + B, "==", 0, eps)
            and compare_numerically(2 * (A + eta) + zeta, ">", 0, eps)
        ):
            if verbose:
                print(f"8 {phrase} {summary_line()}")
            C = A + B + C + xi + eta + zeta
            xi = 2 * B + xi + zeta
            eta = 2 * A + eta + zeta
            # go to 1
            continue
        break
    if verbose:
        cprint(
            f"result:    {summary_line()}",
            color="green",
        )
    if return_cell:
        a = sqrt(A)
        b = sqrt(B)
        c = sqrt(C)
        alpha = acos(xi / 2 / b / c) * TODEGREES
        beta = acos(eta / 2 / a / c) * TODEGREES
        gamma = acos(zeta / 2 / a / b) * TODEGREES
        return a, b, c, alpha, beta, gamma
    return np.array([[A, B, C], [xi / 2, eta / 2, zeta / 2]])


def check_cub(angles: np.ndarray, axes: np.ndarray, eps):
    target_angles = np.array(
        [
            [0, 45, 45, 45, 45, 90, 90, 90, 90],
            [0, 45, 45, 45, 45, 90, 90, 90, 90],
            [0, 45, 45, 45, 45, 90, 90, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
            [0, 45, 45, 60, 60, 60, 60, 90, 90],
        ]
    )

    conventional_axis = np.array([0, 45, 45, 45, 45, 90, 90, 90, 90])

    axes = np.array([i[0] for i in axes])

    if 9 <= angles.shape[0]:
        sub_angles = angles[:9, :9]
        sub_axes = axes[:9]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < eps
        ).all():
            xyz = []
            for i in range(9):
                if (np.abs(np.sort(sub_angles[i]) - conventional_axis) < eps).all():
                    xyz.append(sub_axes[i])
            det = np.abs(np.linalg.det(xyz))
            if det == 1:
                result = "CUB"
            elif det == 4:
                result = "FCC"
            elif det == 2:
                result = "BCC"
            return result, False
        return None, True
    return None, True


def check_hex(angles: np.ndarray, eps):
    target_angles = np.array(
        [
            [0, 90, 90, 90, 90, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
            [0, 30, 30, 60, 60, 90, 90],
        ]
    )
    angles = angles[:7]
    if 7 <= angles.shape[0]:
        sub_angles = angles[:7, :7]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < eps
        ).all():
            return "HEX", False
        return None, True
    return None, True


def check_tet(angles: np.ndarray, axes: np.ndarray, eps, cell):
    target_angles = np.array(
        [
            [0, 90, 90, 90, 90],
            [0, 45, 45, 90, 90],
            [0, 45, 45, 90, 90],
            [0, 45, 45, 90, 90],
            [0, 45, 45, 90, 90],
        ]
    )

    conventional_axis = np.array([0, 90, 90, 90, 90])

    axes = np.array([i[0] for i in axes])

    angles = angles[:5]
    if 5 <= angles.shape[0]:
        sub_angles = angles[:5, :5]
        sub_axes = axes[:5]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < eps
        ).all():
            xy = []
            for i in range(5):
                if (np.abs(np.sort(sub_angles[i]) - conventional_axis) < eps).all():
                    z = sub_axes[i]
                else:
                    xy.append(sub_axes[i])
            xy.sort(key=lambda x: np.linalg.norm(x @ cell))

            xyz = [z, xy[0], xy[1]]

            det = np.abs(np.linalg.det(xyz))
            if det == 1:
                result = "TET"
            elif det == 2:
                result = "BCT"
            return result, False
        return None, True
    return None, True


def check_rhl(angles: np.ndarray, eps):
    target_angles = np.array(
        [
            [0, 60, 60],
            [0, 60, 60],
            [0, 60, 60],
        ]
    )

    angles = angles[:3]
    if 3 <= angles.shape[0]:
        sub_angles = angles[:3, :3]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < eps
        ).all():
            return "RHL", False
        return None, True
    return None, True


def check_orc(angles: np.ndarray, axes: np.ndarray, eps):
    target_angles = np.array(
        [
            [0, 90, 90],
            [0, 90, 90],
            [0, 90, 90],
        ]
    )

    angles = angles[:3]
    axes = np.array([i[0] for i in axes])
    if 3 <= angles.shape[0]:
        sub_angles = angles[:3, :3]
        sub_axes = axes[:3]
        if (
            np.abs(np.sort(sub_angles.flatten()) - np.sort(target_angles.flatten()))
            < eps
        ).all():
            C = np.array(sub_axes, dtype=float).T
            det = np.abs(np.linalg.det(C))
            if det == 1:
                result = "ORC"
            if det == 4:
                result = "ORCF"
            if det == 2:
                v = C @ [1, 1, 1]

                def gcd(p, q):
                    while q != 0:
                        p, q = q, p % q
                    return p

                if (
                    gcd(abs(v[0]), abs(v[1])) > 1
                    and gcd(abs(v[0]), abs(v[2])) > 1
                    and gcd(abs(v[1]), abs(v[2])) > 1
                ):
                    result = "ORCI"
                else:
                    result = "ORCC"
            return result, False
        return None, True
    return None, True


def get_perpendicular_shortest(v, cell, eps):
    perp_axes = []

    miller_indices = (np.indices((3, 3, 3)) - 1).transpose((1, 2, 3, 0)).reshape(27, 3)

    for index in miller_indices:
        if (index != [0, 0, 0]).any():
            if abs((index @ cell) @ (v @ cell)) < eps:
                perp_axes.append(index)

    perp_axes.sort(key=lambda x: np.linalg.norm(x @ cell))

    # indices 0 and 2 (not 0 and 1), since v and -v are present in miller_indices
    return perp_axes[0], perp_axes[2]


def check_mcl(angles: np.ndarray, axes: np.ndarray, eps, cell):
    axes = np.array([i[0] for i in axes])
    angles = angles[:1]
    if 1 <= angles.shape[0]:
        b = axes[0]

        a, c = get_perpendicular_shortest(b, cell, eps)

        C = np.array(
            [
                a,
                b,
                c,
            ],
            dtype=float,
        ).T
        det = np.abs(np.linalg.det(C))
        if det == 1:
            return "MCL", False
        if det == 2:
            return "MCLC", False
        return None, True
    return None, True


def lepage(
    a=1,
    b=1,
    c=1,
    alpha=90,
    beta=90,
    gamma=90,
    eps_rel=1e-4,
    verbose=False,
    very_verbose=False,
    give_all_results=False,
    delta_max=0.01,
):
    r"""
    Le Page algorithm [1]_.

    Parameters
    ----------
    a : float, default 1
        Length of the :math:`a_1` vector.
    b : float, default 1
        Length of the :math:`a_2` vector.
    c : float, default 1
        Length of the :math:`a_3` vector.
    alpha : float, default 90
        Angle between vectors :math:`a_2` and :math:`a_3`. In degrees.
    beta : float, default 90
        Angle between vectors :math:`a_1` and :math:`a_3`. In degrees.
    gamma : float, default 90
        Angle between vectors :math:`a_1` and :math:`a_2`. In degrees.
    eps_rel : float, default 1e-4
        Relative epsilon.
    verbose : bool, default False
        Whether to print the steps of an algorithm.
    very_verbose : bool, default False
        Whether to print the detailed steps of an algorithm.
    give_all_results : bool, default False
        Whether to give the whole list of consecutive results.
    delta_max : float, default 0.01
        Maximum angle tolerance, in degrees.

    Returns
    -------
    result : str
        Bravais lattice type. If give_all_results == True, then return list of all
        consecutive results.

    References
    ----------
    .. [1] Le Page, Y., 1982.
        The derivation of the axes of the conventional unit cell from
        the dimensions of the Buerger-reduced cell.
        Journal of Applied Crystallography, 15(3), pp.255-259.
    """

    if very_verbose:
        verbose = True

    if volume(a, b, c, alpha, beta, gamma) == 0:
        raise ValueError("Cell volume is zero")

    limit = max(1.5, delta_max * 1.1)
    eps_volumetric = eps_rel * volume(a, b, c, alpha, beta, gamma) ** (1 / 3.0)
    decimals = abs(floor(log10(abs(eps_volumetric))))
    if delta_max is None:
        delta_max = eps

    # Niggli reduction
    try:
        a, b, c, alpha, beta, gamma = niggli(
            a, b, c, alpha, beta, gamma, return_cell=True
        )
    except:
        import warnings

        warnings.warn(
            "LePage algorithm: Niggli reduction failed, using input parameters",
            RuntimeWarning,
        )

    cell = Cell.from_params(a, b, c, alpha, beta, gamma)
    rcell = Cell.reciprocal(cell)
    if very_verbose:
        print("Cell:")
        print_2d_array(cell, fmt=f"{4+decimals}.{1+decimals}f")
        print("Reciprocal cell:")
        print_2d_array(rcell, fmt=f"{4+decimals}.{1+decimals}f")

    # Find all axes with twins
    miller_indices = (np.indices((5, 5, 5)) - 2).transpose((1, 2, 3, 0)).reshape(125, 3)
    axes = []
    for U in miller_indices:
        for h in miller_indices:
            if abs(U @ h) == 2:
                t = U @ cell
                tau = h @ rcell
                delta = (
                    np.arctan(np.linalg.norm(np.cross(t, tau)) / abs(t @ tau))
                    * TODEGREES
                )
                if delta < limit:
                    axes.append([U, t / np.linalg.norm(t), abs(U @ h), delta])

    # Sort and filter
    axes.sort(key=lambda x: x[-1])
    keep_index = np.ones(len(axes))
    for i in range(len(axes)):
        if keep_index[i]:
            for j in range(i + 1, len(axes)):
                if (
                    (axes[i][0] == axes[j][0]).all()
                    or (axes[i][0] == -axes[j][0]).all()
                    or (axes[i][0] == 2 * axes[j][0]).all()
                    or (axes[i][0] == 0.5 * axes[j][0]).all()
                ):
                    keep_index[i] = 0
                    break
    new_axes = []
    for i in range(len(axes)):
        if keep_index[i]:
            if set(axes[i][0]) == {0, 2}:
                axes[i][0] = axes[i][0] / 2
            new_axes.append(axes[i])
    axes = new_axes

    if very_verbose:
        print(f"Axes with delta < {limit}:")
        print(f"       U     {'delta':>{4+decimals}}")
        for ax in axes:
            print(
                f"  ({ax[0][0]:2.0f} "
                + f"{ax[0][1]:2.0f} "
                + f"{ax[0][2]:2.0f}) "
                + f"{ax[-1]:{4+decimals}.{1+decimals}f}"
            )

    # Compute angles matrix
    n = len(axes)
    angles = np.zeros((n, n), dtype=float)
    for i in range(n):
        for j in range(i, n):
            angles[i][j] = (
                acos(abs(np.clip(np.array(axes[i][1]) @ np.array(axes[j][1]), -1, 1)))
                * TODEGREES
            )
    angles += angles.T

    # Main check cycle
    delta = None
    separator = lambda x: "=" * 20 + f" Cycle {x} " + "=" * 20
    cycle = 0
    if give_all_results:
        results = []
    while delta is None or delta > delta_max:
        if verbose:
            cycle += 1
            print(separator(cycle))

        try:
            delta = max(axes, key=lambda x: x[-1])[-1]
        except ValueError:
            delta = 0
        eps = max(eps_volumetric, delta)
        decimals = abs(floor(log10(abs(eps))))
        if very_verbose:
            decimals = abs(floor(log10(abs(eps))))
            print(
                f"Epsilon = {eps:{4+decimals}.{1+decimals}f}\n"
                + f"delta   = {delta:{4+decimals}.{1+decimals}f}"
            )
            print("Axes:")
            print(f"       U     {'delta':>{4+decimals}}")
            for ax in axes:
                print(
                    f"  ({ax[0][0]:2.0f} "
                    + f"{ax[0][1]:2.0f} "
                    + f"{ax[0][2]:2.0f}) "
                    + f"{ax[-1]:{4+decimals}.{1+decimals}f}"
                )
            print("Angles:")
            print_2d_array(angles, fmt=f"{4+decimals}.{1+decimals}f")

        continue_search = True
        n = len(axes)
        result = None

        # CUB
        result, continue_search = check_cub(angles, axes, eps)

        # HEX
        if continue_search:
            result, continue_search = check_hex(angles, eps)

        # TET
        if continue_search:
            result, continue_search = check_tet(angles, axes, eps, cell)

        # RHL
        if continue_search:
            result, continue_search = check_rhl(angles, eps)

        # ORC
        if continue_search:
            result, continue_search = check_orc(angles, axes, eps)

        # MCL
        if continue_search:
            result, continue_search = check_mcl(angles, axes, eps, cell)

        # TRI
        if continue_search:
            result = "TRI"

        if verbose:
            print(
                f"System {result} with the worst delta = {delta:{4+decimals}.{1+decimals}f}"
            )

        if len(axes) > 0:
            # remove worst axes
            while len(axes) >= 2 and axes[-1][-1] == axes[-2][-1]:
                axes = axes[:-1]
                angles = np.delete(angles, -1, -1)[:-1]
            axes = axes[:-1]
            angles = np.delete(angles, -1, -1)[:-1]

        if give_all_results:
            results.append((result, delta))

    if give_all_results:
        return results

    return result



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
Crystal module describes the structure and its properties.
"""

from . import cell as Cell
from . import constants as crystal_constants
from .atom import Atom
from .bravais_lattice import *
from .constants import *
from .crystal import Crystal
from .identify import *
from .kpoints import *
from .lattice import *
from .lattice_plotter import *
from .properties import *

__all__ = ["Atom", "Crystal", "Cell", "crystal_constants"]
__all__.extend(bravais_lattice.__all__)
__all__.extend(lattice.__all__)
__all__.extend(kpoints.__all__)
__all__.extend(identify.__all__)
__all__.extend(properties.__all__)
__all__.extend(lattice_plotter.__all__)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""
The module provides input-output routines.
It does not mean to absorb all interfaces to the external data formats,
but designed to be the place for the constructor of the internal
data structures from the input data of the external programs,
as well as from the internal-specified formats.
"""

from .internal import *
from .tb2j import *
from .vampire import *
from .vasp import *

__all__ = []
__all__.extend(internal.__all__)
__all__.extend(tb2j.__all__)
__all__.extend(vampire.__all__)
__all__.extend(vasp.__all__)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
import numpy as np

from radtools.decorate.axes import plot_hlines, plot_vlines
from radtools.dos.pdos import PDOS

__all__ = ["COLOURS", "plot_projected"]

COLOURS = [
    "#00FFFF",
    "#FF9720",
    "#CD00FF",
    "#FFFF2B",
    "#00B9FF",
    "#FF163D",
    "#79FF35",
    "#FF0BEA",
    "#0200FF",
]


def plot_projected(
    pdos: PDOS,
    efermi=0.0,
    output_name="pdos",
    title=None,
    xlim=None,
    ylim=None,
    relative=False,
    normalize=False,
    interactive=False,
    save_pickle=False,
    colours=COLOURS,
    total_label="default",
    axes_labels_fontsize=14,
    legend_fontsize=12,
    title_fontsize=18,
):
    r"""
    Plot PDOS.

    Parameters
    ----------
    pdos : :py:class:`.PDOS`
        PDOS for the plot.
    efermi : float, default 0
        Fermi energy.
    output_name : str, default "pdos"
        output_name for the plot file. Extension ".png" is added at the end.
    title : str, optional
        Title of the plot. Passed to the ``ax.set_title()``.
    xlim : tuple
        limits for the x (Energy) axis
    ylim : tuple
        limits for the y (PDOS) axis
    relative : bool, default False
        Relative plot style.
    normalize : bool, default False
        Whether to norma;ize relative plot style.
    interactive : bool, default False
        Whether to use interactive plotting mode.
    save_pickle : bool, default False
        Whether to save figure as a .pickle file.
        Helps for custom modification of particular figures.
    colours : list
        List of colours to be used. values are passed directly to matplotlib
    total_label : str or ``None``, default "default"
        Label for the total data. If None , then the label is not added
    axes_label_fontsize : int, default 14
        Fontsize of the axes labels.
    legend_fontsize : int, default 12
        Fontsize of the legend.
    title_fontsize : int, default 18
        Fontsize of the title.
    """

    n = len(pdos.projectors)
    pdos = pdos.squeezed()

    if relative:
        fig, ax = plt.subplots(figsize=(9, 4))
    else:
        fig, axs = plt.subplots(n, 1, figsize=(9, n * 2))
        if n == 1:
            axs = [axs]
    fig.subplots_adjust(hspace=0)

    def set_up_axis(ax, i):
        if normalize:
            ax.set_ylabel("PDOS / LDOS", fontsize=axes_labels_fontsize)
        else:
            ax.set_ylabel("DOS, states/eV", fontsize=axes_labels_fontsize)
        if i == n - 1:
            if efermi == 0:
                ax.set_xlabel("E, eV", fontsize=axes_labels_fontsize)
            else:
                ax.set_xlabel("E - E$_{Fermi}$, eV", fontsize=axes_labels_fontsize)
        else:
            ax.axes.get_xaxis().set_visible(False)
        if ylim is not None:
            ax.set_ylim(*tuple(ylim))
        if xlim is not None:
            ax.set_xlim(*tuple(xlim))
        else:
            ax.set_xlim(np.amin(pdos.energy), np.amax(pdos.energy))
        if efermi != 0:
            plot_vlines(ax, 0)

        if title is not None and (i == 0 or relative):
            ax.set_title(title, fontsize=title_fontsize)

    if normalize:
        pdos = pdos.normalized(zeros_to_none=False)

    if relative:
        set_up_axis(ax, n - 1)
        plot_hlines(ax, 0)

    for i, projector in enumerate(pdos):
        if not relative:
            ax = axs[i]
            set_up_axis(ax, i)
        if pdos.spin_pol:
            if total_label == "default":
                label_up = f"{pdos.projectors_group} (up)"
                label_down = f"{pdos.projectors_group} (down)"
            elif total_label is None:
                label_up = None
                label_down = None
            else:
                label_up = f"{total_label} (up)"
                label_down = f"{total_label} (down)"
            if relative:
                if i == 0 and total_label is not None:
                    ax.plot(
                        pdos.energy,
                        pdos.ldos[0],
                        "-",
                        lw=1,
                        color="blue",
                        alpha=0.8,
                        label=label_up,
                    )
                    ax.plot(
                        pdos.energy,
                        -pdos.ldos[1],
                        "-",
                        lw=1,
                        color="red",
                        alpha=0.8,
                        label=label_down,
                    )
                ax.fill_between(
                    pdos.energy,
                    np.sum(pdos[:i], axis=0)[0],
                    np.sum(pdos[: i + 1], axis=0)[0],
                    lw=0,
                    color=colours[i % len(colours)],
                    # alpha=0.5,
                    label=f"{projector}",
                )
                ax.fill_between(
                    pdos.energy,
                    -np.sum(pdos[:i], axis=0)[1],
                    -np.sum(pdos[: i + 1], axis=0)[1],
                    lw=0,
                    color=colours[i % len(colours)],
                    # alpha=0.5,
                )
            else:
                if total_label is not None:
                    ax.fill_between(
                        pdos.energy,
                        0,
                        pdos.ldos[0],
                        lw=0,
                        color="blue",
                        alpha=0.2,
                        label=label_up,
                    )
                    ax.fill_between(
                        pdos.energy,
                        0,
                        -pdos.ldos[1],
                        lw=0,
                        color="red",
                        alpha=0.2,
                        label=label_down,
                    )

                ax.plot(
                    pdos.energy,
                    pdos[projector][0],
                    "-",
                    lw=0.8,
                    color="blue",
                    alpha=0.8,
                    label=f"{projector} (up)",
                )
                ax.plot(
                    pdos.energy,
                    -pdos[projector][1],
                    "-",
                    lw=0.8,
                    color="red",
                    alpha=0.8,
                    label=f"{projector} (down)",
                )
        else:
            if total_label == "default":
                total_label = pdos.projectors_group
            if relative:
                if i == 0 and total_label is not None:
                    ax.plot(
                        pdos.energy,
                        pdos.ldos,
                        "-",
                        lw=1,
                        color="black",
                        alpha=0.8,
                        label=total_label,
                    )
                ax.fill_between(
                    pdos.energy,
                    np.sum(pdos[:i], axis=0),
                    np.sum(pdos[: i + 1], axis=0),
                    lw=0,
                    color=colours[i % len(colours)],
                    # alpha=0.5,
                    label=projector,
                )

            else:
                if total_label is not None:
                    ax.fill_between(
                        pdos.energy,
                        0,
                        pdos.ldos,
                        lw=0,
                        color="black",
                        alpha=0.3,
                        label=total_label,
                    )
                ax.plot(
                    pdos.energy,
                    pdos[projector],
                    "-",
                    lw=0.8,
                    color="black",
                    alpha=0.8,
                    label=projector,
                )
        if interactive:
            ax.legend(
                loc=(1.025, 0.2),
                bbox_transform=ax.transAxes,
                draggable=True,
                fontsize=legend_fontsize,
            )
        else:
            ax.legend(
                loc=(1.025, 0.2), bbox_transform=ax.transAxes, fontsize=legend_fontsize
            )

    if interactive:
        plt.show()
    else:
        plt.savefig(f"{output_name}.png", dpi=600, bbox_inches="tight")
        if save_pickle:
            import pickle

            with open(f"{output_name}.pickle", "wb") as file:
                pickle.dump(fig, file)
    plt.close()



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from matplotlib.colors import LinearSegmentedColormap, to_rgb

__all__ = ["custom_cmap"]


def custom_cmap(start_color, finish_color):
    r"""
    Prepare custom colormap. From one color to the other.

    Parameters
    ----------
    start_color : Color
        Start color.
    finish_color : Color
        Finish color.
    """

    r1, g1, b1 = to_rgb(start_color)

    r2, g2, b2 = to_rgb(finish_color)

    cdict = {
        "red": ((0, r1, r1), (1, r2, r2)),
        "green": ((0, g1, g1), (1, g2, g2)),
        "blue": ((0, b1, b1), (1, b2, b2)),
    }

    return LinearSegmentedColormap("custom_cmap", cdict, N=256)



# RAD-tools - program for spin Hamiltonian and magnons.
# Copyright (C) 2022-2024  Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from argparse import ArgumentParser
from calendar import month_name
from datetime import datetime

import numpy as np
from termcolor import cprint

from radtools import __version__ as version
from radtools.decorate.stats import logo
from radtools.io.tb2j import load_tb2j_model


def manager(
    output_name="template.txt",
    input_filename=None,
    R_vector=None,
    max_distance=None,
    min_distance=None,
    distance=None,
    verbose=False,
    eps=1e-3,
):
    r"""
    :ref:`rad-make-template` script.

    Full documentation on the behaviour is available in the
    :ref:`User Guide <rad-make-template>`.
    Parameters of the function directly
    correspond to the arguments of the script.

    Parameters
    ----------
    output_name : str, default "template.txt"
        Name for the template output file.

        See also: :ref:`example <output-notes>`.

        Console argument: ``-on`` / ``--output-name``

        Metavar: "filename"
    input_filename : str, optional
        Relative or absolute path to the "exchange.out" file, including name and extension of the file.

        Console argument: ``-if`` / ``--input-filename``

        Metavar: "filename"

        .. versionchanged:: 0.5.12 Renamed from "tb2j_filename"
    R_vector : list of int, optional
        R vectors for filtering the spin Hamiltonian.

        In TB2J outputs the bond is defined by atom 1 (from) and atom 2 (to).
        Atom 1 is always located in (0, 0, 0) unit cell, while atom 2 is located in
        R = (i, j, k) unit cell. This parameter tells the script to keep only the
        bonds for which atom 2 is located in one of specified R supercells.
        Supercells are specified by a set of integers separated by spaces.
        They are grouped by three starting from the left and forms a set
        of R vectors. If the last group contains 1 or 2 integers they are ignored.

        Console argument: ``-R`` / ``--R-vector``

        Metavar: "i1 j1 k1 i2 j2 k2 ..."
    max_distance : float, optional
        (<=) Maximum distance.

        All the bonds with the distance between atom 1 and atom 2
        greater than maximum distance are excluded from the model.

        Console argument: ``-maxd`` / ``--max-distance``

        Metavar: "distance"
    min_distance : float, optional
        (>=) Minimum distance.

        All the bonds with the distance between atom 1 and atom 2
        lower than minimum distance are excluded from the Hamiltonian.

        Console argument: ``-mind`` / ``--min-distance``

        Metavar: "distance"
    distance : float, optional
        (=) Exact distance.

        Only the bonds with the exact distance remains in the model.
        There is no point in specifying maximum or minimum distance when
        this parameter is provided.

        Console argument: ``-d`` / ``--distance``

        Metavar: "distance"
    verbose : bool, default False
        Verbose output, propagates to the called methods.

        Console argument: ``-v`` / ``--verbose``
    eps : float, default 1e-3
        Epsilon for the distance comparison.

        Console argument: ``--eps``
    """

    n_sep = 80

    out_head, out_tail = os.path.split(output_name)
    if len(out_tail) == 0:
        out_tail = "template.txt"

    output_name = os.path.join(out_head, out_tail)

    # Create the output directory if it does not exist
    if out_head != "":
        os.makedirs(out_head, exist_ok=True)

    # Get current date and time
    cd = datetime.now()

    if distance is not None:
        min_distance = distance
        max_distance = distance

    # Translate sequence of numbers to R vectors
    if R_vector is not None:
        R_vector = np.array(R_vector[: len(R_vector) // 3 * 3], dtype=int).reshape(
            (len(R_vector) // 3, 3)
        )
        R_vector = list(map(tuple, R_vector.tolist()))

    # Template draft
    template = (
        "=" * n_sep
        + "\n"
        + "Neighbors template:\n"
        + f"Atom1 Atom2 {'i':>3} {'j':>3} {'k':>3}\n"
        + "-" * n_sep
        + "\n"
        + "J1 $J_1$\n"
        + f"atom1 atom2 {0:3} {0:3} {0:3}\n"
        + f"atom1 atom2 {1:3} {0:3} {0:3}\n"
        + f"atom1 atom1 {-1:3} {0:3} {2:3}\n"
        + "-" * n_sep
        + "\n"
        + "J2\n"
        + f"atom2 atom1 {9:3} {5:3} {-3:3}\n"
        + f"atom1 atom2 {1:3} {4:3} {0:3}\n"
        + f"atom2 atom2 {1:3} {0:3} {2:3}\n"
        + "=" * n_sep
        + "\n"
    )
    with open(output_name, "w", encoding="utf-8") as file:
        file.write("=" * n_sep + "\n" + logo(date_time=True, line_length=n_sep) + "\n")
        # Write the draft
        if input_filename is None:
            file.write(template)
        # Create the template based on the input file
        else:
            file.write(
                f"Template is created based on the file:\n{os.path.abspath(input_filename)}\n"
            )

            # Read and filter the model
            model = load_tb2j_model(input_filename, quiet=not verbose)
            model.filter(
                min_distance=min_distance, max_distance=max_distance, R_vector=R_vector
            )

            # Write header
            file.write(
                "=" * n_sep
                + "\n"
                + "Neighbors template:\n"
                + f"Atom1 Atom2 {'i':>3} {'j':>3} {'k':>3}\n"
                + "-" * n_sep
                + "\n"
            )

            # Get bonds from the model
            data = []
            for atom1, atom2, R, J in model:
                data.append(
                    (
                        atom1.name,
                        atom2.name,
                        R,
                        model.get_distance(atom1, atom2, R),
                    )
                )

            # Sort bonds by distance
            data.sort(key=lambda x: x[3])

            j = 1
            file.write(f"J{j} " + "$J_{" + f"{j}" + "}$\n")
            file.write(
                f"{data[0][0]:5} {data[0][1]:5} "
                + f"{data[0][2][0]:3.0f} {data[0][2][1]:3.0f} {data[0][2][2]:3.0f}\n"
            )
            for i, (atom1, atom2, R, distance) in enumerate(data[1:]):
                # If distance is the same as the previous one, write the bond
                if abs(distance - data[i][3]) < eps:
                    file.write(
                        f"{atom1:5} {atom2:5} "
                        + f"{R[0]:3.0f} {R[1]:3.0f} {R[2]:3.0f}\n"
                    )
                # If distance is different, start a new group and write the bond
                else:
                    j += 1
                    file.write("-" * n_sep + "\n")
                    file.write(f"J{j} " + "$J_{" + f"{j}" + "}$\n")
                    file.write(
                        f"{atom1:5} {atom2:5} "
                        + f"{R[0]:3.0f} {R[1]:3.0f} {R[2]:3.0f}\n"
                    )

            file.write("=" * n_sep + "\n")
    cprint(
        f"Template draft is in "
        + f"{os.path.abspath(output_name)}, grouped by distance",
        "blue",
    )
    cprint(f"Do not forget to correct the template draft to your needs!", "yellow")


def create_parser():
    parser = ArgumentParser()
    parser.add_argument(
        "-on",
        "--output-name",
        default="template.txt",
        metavar="filename",
        type=str,
        help="Name for the template output file.",
    )
    parser.add_argument(
        "-if",
        "--input-filename",
        default=None,
        metavar="filename",
        type=str,
        help='Relative or absolute path to the "exchange.out" file, including name and extension of the file.',
    )
    parser.add_argument(
        "-R",
        "--R-vector",
        default=None,
        metavar="i1 j1 k1 i2 j2 k2 ...",
        type=int,
        nargs="*",
        help="R vectors for filtering the spin Hamiltonian.",
    )
    parser.add_argument(
        "-maxd",
        "--max-distance",
        default=None,
        metavar="distance",
        type=float,
        help="(<=) Maximum distance.",
    )
    parser.add_argument(
        "-mind",
        "--min-distance",
        default=None,
        metavar="distance",
        type=float,
        help="(>=) Minimum distance.",
    )
    parser.add_argument(
        "-d",
        "--distance",
        default=None,
        metavar="distance",
        type=float,
        help="(=) Exact distance.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Verbose output, propagates to the called methods.",
    )
    parser.add_argument(
        "--eps",
        default=1e-3,
        type=float,
        help="Epsilon for the distance comparison.",
    )

    return parser



#! /usr/local/bin/python3

from argparse import ArgumentParser

import matplotlib.pyplot as plt
import numpy as np
import yaml


def plot(file, kp=None, constant=1, on="output"):
    data_file = open(file, "r")
    data = yaml.load(data_file, Loader=yaml.FullLoader)
    data_file.close()
    X = []
    Y = []
    K_points = []
    miny = None
    maxy = None
    for i, point in enumerate(data["phonon"]):
        for j, band in enumerate(data["phonon"][i]["band"]):
            data["phonon"][i]["band"][j]["frequency"] *= constant
    for i, point in enumerate(data["phonon"]):
        print(i)
        X.append(point["distance"])
        Y.append(list(map(lambda x: x["frequency"], point["band"])))
        if (
            i == 0
            or i == len(data["phonon"]) - 1
            or data["phonon"][i]["q-position"] == data["phonon"][i + 1]["q-position"]
        ):
            K_points.append(point["distance"])
        if not miny:
            miny = min(Y[-1])
        if not maxy:
            maxy = max(Y[-1])
        miny = min(min(Y[-1]), miny)
        maxy = max(max(Y[-1]), maxy)

    print(K_points)

    fig, ax = plt.subplots()
    X = np.array(X)
    Y = np.array(Y)
    ax.plot(X, Y, color="red")
    ax.set_ylabel("Phonon frequency (THz)", fontsize=15)
    ax.set_xlim(X[0], X[-1])
    if kp is None or len(kp) != len(K_points):
        if kp is None:
            kp = []
        print(
            f"\033[93m"
            f"{len(K_points)} k-points expected, only {len(kp)} k-points are provided. "
            f"Please, check the input."
            f"\nProvided K-points: {kp}"
            f"\x1b[0m"
        )
    if kp and len(kp) == len(K_points):
        for i, k in enumerate(kp):
            if k == "G":
                kp[i] = "$\Gamma$"
        ax.xaxis.set_ticks(K_points, kp, fontsize=15)
        for k_point in K_points:
            ax.vlines(
                k_point,
                0,
                1,
                transform=ax.get_xaxis_transform(),
                color="black",
                linewidths=0.5,
            )
    plt.show()
    if constant != 1:
        output = yaml.dump(data)
        with open(f"{on}.yaml", "w", encoding="utf-8") as file:
            file.write(output)


if __name__ == "__main__":
    parser = ArgumentParser(description="Fit Exchange parameters map")

    parser.add_argument(
        "-file", type=str, default="band.yaml", help="Path to the band.yaml file"
    )
    parser.add_argument(
        "-kp", type=str, default=None, nargs="*", help="K-path, example: -kp G X R Y G"
    )
    parser.add_argument(
        "-constant",
        type=float,
        default=1,
        help="Constant for conversion. " "The values will be multiplied by it",
    )
    parser.add_argument("-on", type=str, default="output", help="output name file")
    args = parser.parse_args()
    if args.file:
        plot(args.file, kp=args.kp, constant=args.constant, on=args.on)
